#!/usr/bin/perl
use strict;
use Data::Dumper;
use Getopt::Long;
use Pod::Usage;
use vars qw(%OPT);
use Storable;
use lib "/home/mapper/lymphoma/code/modules";
use Set::IntSpan;
use Benchmark;

# store the arguments passed in, to write to log file
my $batch_cmd = join " ", @ARGV;

GetOptions(\%OPT,
     "input_dir=s",
     "outdir=s",
     "adj",
     "main", # REMINDER: this should probably be default
     "min_k=i",
     #"bam=s", # REMINDER: do not use reads to genome bam file
     "only_blat_aligns",
     "only_exon_aligns",
     "require_both_aligners",
     "max_match_fraction=f",
     "min_merge_overlap=f",
     "min_ctg_cover_fraction=f",
     "num_aligns=i",
     "min_identity=f",
     "ctg_report_min=i",
     "read_pair_trump=i", # REMINDER: DO NOT USE THIS UNTIL WE FIGURE OUT PHYS COV
     "min_read_coverage=i",
     "min_source_num=i",
     "nosingle_k",
     "nolocal",
     "noRNA",
     "noMito",
     "run_kaligner",
     "read_file_list=s",
     "read_seq_dir=s",
     "read_length=i",
     "start=i",
     "run_acf",
     "use_gap_filter",
     "skip_gap_filter",
     "min_gap_size=i",
     "min_gap_similarity=f",
     "debug",
     "help|h",
     "man|m",
     "ignore_cmds=s"
     );
pod2usage(-verbose => 2) if $OPT{man};
pod2usage(1) if ($OPT{help} || !$OPT{input_dir});

=pod

=head1 SYNOPSIS

batch_fac_filter_and_support_1.0.pl -input_dir DIR [ OPTIONS ]

NOTE: DIR should be the base directory for the library assembly

=head1 OPTIONS

 -outdir DIR        Output_directory [ default: INPUT_DIR/anomalous_contigs ]
 -adj               Run on merged adjacency contigs
 -main              Run on merged main contigs
 -min_k N           Do not consider contigs assembled with k lower than N
 -only_blat_aligns  Only run on blat alignments
 -only_exon_aligns  Only run on exon alignments
 -require_both_aligners
                    Require evidence from both aligners
 -max_match_fraction F
                    If any single alignment for a contig uses at least F%
                    of the contig length, do not look for any split
                    alignments for that contig. [ default: 0.95 ]
 -min_merge_overlap F
                    If at least F% of two alignments represent the same
                    region of the contig, consider only the higher scoring
                    of the two alignments. [ default: 0.85 ]
 -min_ctg_cover_fraction F
                    Only consider split alignments for which the two
                    alignments represent at least F% of the total contig.
                    [ default: 0.85 ]
 -num_aligns N      Extract the best N alignments for each contig from the
                    alignments file. [ default: 10 ]
 -min_identity PID  Extract from the alignments file only alignments with a
                    percent identity of at least PID%. [ default: 90.0 ]
 -ctg_report_min N  At least one contig showing the event must appear in no
                    more than N events
 -read_pair_trump N
                    Pass any event with an average physical coverage of at
                    least N in the event region
 -min_read_coverage N
                    Require at least N_reads_aligning_to the event region
 -min_source_num N  Require at least N sources to show the event (sources
                    include adjacency contigs, main contigs, and different
                    k values)
 -nosingle_k        Require multiple contigs from multiple k values to
                    show the event
 -nolocal           Require that event does not represent a local alignment
 -noRNA             Fail events overlapping RNA repeats
 -noMito            Fail events occuring purely in mitochondrial DNA
 -run_kaligner      Run kalign to align the reads to the contigs
 -read_file_list FILE
                    FILE contains paths to the files containing the read
                    sequences used to assemble the aligned contigs
 -read_seq_dir DIR  The read sequence files are found in directory DIR
 -read_length N     Each read is N bp long [ default: 50 ]
 -run_acf           Run the anomalous contig finder to identify anomalous
                    contigs by examining their alignments
 -use_gap_filter    Search for interesting gapped BLAT alignments
 -skip_gap_filter   Do not search for interesting gapped BLAT alignments
 -min_gap_size N    Only consider gaps of size N or greater [ default: 20 ]
 -min_gap_similarity
                    Require that gap sequences map to the rest of the
                    contig with at least F% identity [ default: 0.95 ]
 -debug             Print status info while running
 -h, -help          Display a brief help message
 -man               Display the full documentation

=head1 NAME

batch_fac_filter_and_support_1.0.pl -> Wrapper to generate the candidate split alignment contigs from blat/exonerate alignments

=head1 DESCRIPTION

Dec 1/09

Script that runs all the anomalous contig code for a library and generates a single summary file with a PASS/FAIL status

=head1 AUTHOR

Matthew Field (edited by Lucas Swanson)

=head1 EXAMPLE

/home/lswanson/my_code/svn/trans-abyss/trunk/analysis/anomalous_contig_finder/batch_fac_filter_and_support_1.0.pl
 -input_dir /projects/validations/transcriptome/HS1184/Assembly/abyss-1.0.13
 -min_k 32 -run_acf

/home/lswanson/my_code/svn/trans-abyss/trunk/analysis/anomalous_contig_finder/batch_fac_filter_and_support_1.0.pl
 -input_dir /projects/validations/transcriptome/HS1205/Assembly/abyss-1.0.13
 -main -adj -require_both_aligners -nosingle_k -min_source_num 1 -nolocal -min_read_coverage 2
 -noRNA -read_pair_trump 5 -ctg_report_min 2 -run_acf_kaligner

=cut

my $t0 = new Benchmark;

# Split Finder Exit Statuses
my $SF_SUCCESS = 0;
my $SF_NO_ALIGNMENTS = 3;
# Anomalous contig finder output columns
my $ACF_CTG_ID     = 0;
my $ACF_TYPE       = 1;
my $ACF_TARGET     = 2;
my $ACF_CTG_COORDS = 3;
my $ACF_JUNCTION   = 4;
my $ACF_META_DATA  = 5;
my $ACF_BLOCKS     = 6;
# Sort command field indices
my $SC_ALIGNER    = 5;
my $SC_CTG_ID     = 6;
my $SC_TYPE       = 7;
my $SC_TARGET     = 8;
my $SC_CTG_COORDS = 9;
my $SC_JUNCTION   = 10;
my $SC_META_DATA  = 11;
my $SC_BLOCKS     = 12;

# setup input and output directories
my $input_dir = $OPT{input_dir};
my $outdir = defined $OPT{outdir}?$OPT{outdir}:"$input_dir/anomalous_contigs";
chomp $outdir;
$outdir =~ s/\/$//; # remove final slash(/) if necessary
mkdir($outdir) if !-d $outdir && $OPT{run_acf};

# try to determine library from input directory name
# assume library is human
my ($lib) = $input_dir =~ /(HS\d{4})/;
# if that does not work, try mouse
($lib) = $input_dir =~ /(MM\d{4})/ if $lib !~ /\w/;
# if that does not work, throw an error
if ($lib !~ /\w/) {
  print "Cannot determine library type (mouse or human) from $input_dir\n";
  exit;
}

# open log file
my $log_file_name = "$outdir/${lib}.anomalous_contig.log";
open(my $logref,">",$log_file_name) ||
  die "Can't open log file $log_file_name $!\n";
&print_full($logref, "PARAMETERS USED: $0 $batch_cmd\n\n");

# SETUP DEFAULT FILTER VALUES
# anomalous contig finder parameters
my $num_aligns = defined $OPT{num_aligns}?$OPT{num_aligns}:10;
my $min_identity = defined $OPT{min_identity}?$OPT{min_identity}:95.0;
my $min_ctg_cover_fraction =
  defined $OPT{min_ctg_cover_fraction}?$OPT{min_ctg_cover_fraction}:0.90;
#Want to allow these 'local' alignments in to use as filters later on
my $min_merge_overlap =
  defined $OPT{min_merge_overlap}?$OPT{min_merge_overlap}:0.80;
my $max_match_fraction =
  defined $OPT{max_match_fraction}?$OPT{max_match_fraction}:0.95;

# gap filter parameters
my $use_gap_filter=0;
if (defined $OPT{use_gap_filter}) {
  $use_gap_filter=1;
}
my $min_gap_size = 20;
if (defined $OPT{min_gap_size}) {
  $min_gap_size = $OPT{min_gap_size};
}
my $min_gap_similarity = 0.95;
if (defined $OPT{min_gap_similarity}) {
  $min_gap_similarity = $OPT{min_gap_similarity};
}

# Filter parameter: min number of contig types showing event (main, adj)
#   adds fail reason TOO_FEW_SOURCES
my $min_source_num = defined $OPT{min_source_num}?$OPT{min_source_num}:1;
# Filter parameter: min number of reads mapped to junction region
#   adds fail reason READ_COVERAGE
my $min_read_coverage =
  defined $OPT{min_read_coverage}?$OPT{min_read_coverage}:0;
# Filter parameter: max number of events a contig can be found in to be
# considered good evidence.
#   adds fail reason CTG_REPORT_MIN
my $ctg_report_min = defined $OPT{ctg_report_min}?$OPT{ctg_report_min}:0;

# Create an empty dictionary to hold the input directory names
my %k_dirs = ();
my $merge_dir = "";
# If not running on main or adjacency contigs, look in the individual
# k values and check the min_k flag
if (!$OPT{main} && !$OPT{adj}) {
  opendir(ABYSS, $input_dir) || die "Can't open input directory $!\n";
  %k_dirs = map{$_ => 1} grep {/^k/} readdir ABYSS;
  if ($OPT{min_k}) {
    for my $kdir (keys %k_dirs) {
      my ($file_k) = $kdir =~ /(\d+)/;
      if ($file_k < $OPT{min_k}) {
        delete $k_dirs{$kdir};
      }
    }
  }
} else {
  $merge_dir = "$input_dir/merge";
  if (not -e "$merge_dir") {
    &print_full($logref, "ERROR: Can't find merge directory: $merge_dir\n");
    die;
  }
  if ($OPT{main}) {
    $k_dirs{main} = 1;
  }
  if ($OPT{adj}) {
    $k_dirs{adj} = 1;
  }
}

#Get the alignment files and populate the data structure
my %files = &Find_Files();
my $num_dirs = keys %files;
if (0 == $num_dirs) {
  &print_full($logref, "ERROR: Cannot find alignment files.\n");
  die;
}
# DEBUG TEMP
if ($OPT{debug}) {
  print "# dirs found: $num_dirs\n";
  print "Files:\n";
  for my $dir (keys %files) {
    print "  Dir: $dir\n";
    my %dir_tools = %{$files{$dir}};
    for my $tool (keys %dir_tools) {
      if ("blat" eq $tool) {
        print "    Tool: $tool\n";
        my @tool_files = @{$dir_tools{$tool}};
        for my $file (@tool_files) {
          print "      $file\n";
        }
      } else {
        print "    $tool - $dir_tools{$tool}\n";
      }
    }
  }
  print "\n";
}
#exit;
# DEBUG TEMP

#Run the anomalous contig finder code
my %num_gaps_found = ();
for my $dir (keys %files) {
  $num_gaps_found{$dir} = 0;
}
&Find_Anomalous_Contigs();
#for my $dir (keys %files) {
#&print_full($logref, "Num Gaps: $num_gaps_found{$dir}\n"); # DEBUG
#}

# REMINDER: Now going to use this for grouping by "event target regions"
# Create an empty dictionary for holding target sets
#   ...for holding event information used to group events
my %target_sets = ();

# Create an empty dictionary for holding fasta file names
my %fasta_data = ();
# Create an empty dictionary for holding contigs to reads data
my %contigs_in_events = ();
#Parse the anomalous contig finder's output files
for my $dir (keys %files) {
  #&print_full($logref, "Dir: $dir\n"); # DEBUG
  $fasta_data{$dir}{fasta} = $files{$dir}{contigs};
  #&print_full($logref, "  Blat file: \"$files{$dir}{blat_fac_out}\"\n"); # DEBUG
  if ("" ne $files{$dir}{blat_fac_out}) {
    if (-e $files{$dir}{blat_fac_out}) {
      &Parse_Anomalous_Contigs($files{$dir}{blat_fac_out},'blat',$dir);
    } else {
      die "BLAT file not found: $files{$dir}{blat_fac_out}\n";
    }
  }
  #&print_full($logref, "  Gap file: \"$files{$dir}{gap_fac_out}\"\n"); # DEBUG
  # REMINDER: check that gaps were found
  if ("" ne $files{$dir}{gap_fac_out}) {
    if (-e $files{$dir}{gap_fac_out}) {
      &Parse_Anomalous_Contigs($files{$dir}{gap_fac_out},'blat',$dir);
    } else {
      my $gap_file = "$files{$dir}{gap_fac_out}";
      my $warning = "WARNING: Gap file not found: $gap_file";
      $warning .= "\n  $num_gaps_found{$dir} anomalous contigs found " .
                  "by gap filter.";
      &print_full($logref, "$warning\n");
    }
  }
  #&print_full($logref, "  Exon file: \"$files{$dir}{exon_fac_out}\"\n"); # DEBUG
  if ("" ne $files{$dir}{exon_fac_out}) {
    if (-e $files{$dir}{exon_fac_out}) {
      &Parse_Anomalous_Contigs($files{$dir}{exon_fac_out},'exon',$dir);
    } else {
      die "Exon file not found: $files{$dir}{exon_fac_out}\n";
    }
  }
}
&print_full($logref, "\n");

#Distance to determine whether we're dealing with local alignments
my $local_buffer = 100000;

#Store the contig stats for reporting -> contigs can appear in multiple
# alignment lines
my %ctg_stats = ();
# Create an empty dictionary for holding the groups
my %groups = ();
#Group the events
&Group();

print Dumper \%target_sets if $OPT{start};

#Parse the RNA repeats file and integrate into the data structure
# but only if the noRNA filter is on
if (defined $OPT{noRNA}) {
  &Parse_RNA_Repeats();
}

print Dumper \%groups if $OPT{start};

#Store all the contig files for generating a single fasta for read alignments
my %fasta_files = ();
my @main_fasta = `ls $merge_dir/${lib}-main.fa 2>/dev/null`;
$fasta_files{main} = \@main_fasta if @main_fasta;
my @adj_fasta = `ls $merge_dir/${lib}-adj.fa 2>/dev/null`;
$fasta_files{adj} = \@adj_fasta if @adj_fasta;
my @bubble_fasta = `ls $input_dir/k*/bubbles.fa 2>/dev/null`;
$fasta_files{bubble} = \@bubble_fasta if @bubble_fasta;

# Parameter for KAlign
my $read_length = defined $OPT{read_length}?$OPT{read_length}:50;

# If the min read coverage option is used
my $ctgs_in_each_group;
if (defined $OPT{min_read_coverage}) {
  # Run kaligner if required (if run_kaligner flag is false, this will just
  # find the kalign output files)
  my ($kalign_files) = &Run_KAligner(\%fasta_files);
  my @kalign_files_arr = (keys %{$kalign_files});
  #&print_full($logref, "KAlign output files:\n@kalign_files_arr\n"); ## DEBUG
  # convert array to scalar to get size of array
  my $num_files_found = @kalign_files_arr;
  if (0 == $num_files_found) {
    &print_full($logref,
      "ERROR: Could not find any KAlign files for read support!\n");
    exit;
  }

  ($ctgs_in_each_group) = &Get_Contigs_In_Each_Group();

  #Initialize read support to 0
  for my $group_id (keys %groups) {
    for my $ctg_id (keys %{$ctgs_in_each_group->{$group_id}}) {
      $groups{$group_id}{$ctg_id}{read_count} = 0;
    }
  }

  #Parse the kaligner files
  for my $kalign_file (sort keys %{$kalign_files}) {
    &Parse_KAligner($kalign_file, \%{$ctgs_in_each_group});
  }
  &print_full($logref, "\n");
}

#Summarize the events
&Summary(\%{$ctgs_in_each_group});

#Generate the fasta file with candidate anomalous contigs
&Fasta();

my $t1 = new Benchmark;
my $td = timediff($t1, $t0);
my $ts = timestr($td);
&print_full($logref,"Total running time: $ts\n");

#Find the alignment files
sub Find_Files {
  my %files = ();
  for my $dir (sort keys %k_dirs) {
    #flag to check whether the alignment files exist
    my $proceed = 0;

    # If the "only blat" option is false, find the exonerate alignment files
    if (!$OPT{only_blat_aligns}) {
      my @exon_files = ();
      if ($dir eq 'adj') {
        my $adj_dir = "$merge_dir/cluster/${lib}a-exon/output";
        if (-d $adj_dir) {
          opendir(DIR,$adj_dir) || die "Can't open $adj_dir $!\n";
          my @files = grep {/seq\.\d+\.exon$/} grep {$_ !~ /^\./} readdir DIR;
          for my $file (sort @files) {
            push @exon_files, "$adj_dir/$file";
          }
        } else {
          my $exon_file = join("/",
            "$merge_dir",
            "adj_aligns",
            "hg18",
            "exonerate_2.2",
            "$lib.merge.adj.hg18.e2g.exon");
          if (-e $exon_file) {
            push @exon_files, $exon_file
          }
        }
        if (0 == @exon_files) {
          my $exon_file = join("/",
            "$merge_dir",
            "adj_aligns",
            "mm9",
            "exonerate_2.2",
            "$lib.merge.adj.mm9.e2g.exon");
          if (-e $exon_file) {
            push @exon_files, $exon_file;
          }
        }
      } elsif ($dir eq 'main') {
        my $main_dir = "$merge_dir/cluster/${lib}m-exon/output";
        if (-d $main_dir) {
          opendir(DIR,$main_dir) || die "Can't open $main_dir $!\n";
          my @files = grep {/seq\.\d+\.exon$/} grep {$_ !~ /^\./} readdir DIR;
          for my $file (sort @files) {
            push @exon_files, "$main_dir/$file";
          }
        } else {
          my $exon_file = join("/",
            "$merge_dir",
            "pe_contig_aligns",
            "hg18",
            "exonerate_2.2",
            "$lib.merge.main.hg18.e2g.exon");
          if (-e $exon_file) {
            push @exon_files, $exon_file;
          }
        }
        if (0 == @exon_files) {
          my $exon_file = join("/",
            "$merge_dir",
            "pe_contig_aligns",
            "mm9",
            "exonerate_2.2",
            "$lib.merge.pe_contigs.mm9.e2g.exon");
          if (-e $exon_file) {
            push @exon_files, $exon_file;
          }
        }
      } else {
        &print_full($logref, "$dir is neither \"adj\" nor \"main\"\n");
      }
      if (@exon_files) {
        &print_full($logref,"Found exonerate alignment files\n");
        $files{$dir}{exon} = \@exon_files;
        $proceed = 1;
      } else {
        &print_full($logref, "No exonerate alignments found\n")
      }
    }

    # If the "only exon" option is false, find the blat alignment files
    if (!$OPT{only_exon_aligns}) {
      my @blat_files = ();
      if ($dir eq 'adj') {
        my $adj_dir = "$merge_dir/cluster/${lib}a-blat/output";
        if (-d $adj_dir) {
          opendir(DIR,$adj_dir) || die "Can't open $adj_dir $!\n";
          my @files = grep {/seq\.\d+\.psl$/} grep {$_ !~ /^\./} readdir DIR;
          for my $file (sort @files) {
            push @blat_files, "$adj_dir/$file";
          }
        } else {
          my $blat_file =
            "$merge_dir/adj_aligns/hg18/blat_34/$lib.merge.adj.hg18.psl";
          if (-e $blat_file) {
            push @blat_files, $blat_file;
          }
        }
        # if no blat files were found yet
        if (0 == @blat_files) {
          my $blat_file = "$merge_dir/adj.psl";
          if (-e $blat_file) {
            push @blat_files, $blat_file;
          }
        }
      } elsif ($dir eq 'main') {
        my $main_dir = "$merge_dir/cluster/${lib}m-blat/output";
        if (-d $main_dir) {
          opendir(DIR,$main_dir) || die "Can't open $main_dir $!\n";
          my @files = grep {/seq\.\d+\.psl$/} grep {$_ !~ /^\./} readdir DIR;
          for my $file (sort @files) {
            push @blat_files, "$main_dir/$file";
          }
        } else {
          my $blat_file = join("/",
            "$input_dir/$dir",
            "pe_contig_aligns",
            "hg18/blat_34",
            "$lib.$dir.main.hg18.psl");
          if (-e $blat_file) {
            push @blat_files, $blat_file;
          }
        }
        # if no blat files were found yet
        if (0 == @blat_files) {
          my $blat_file = join("/",
            "$input_dir/$dir",
            "pe_contig_aligns",
            "mm9/blat_34",
            "$lib.$dir.pe_contigs.mm9.psl");
          if (-e $blat_file) {
            push @blat_files, $blat_file;
          }
        }
      } else {
        &print_full($logref, "$dir is neither \"adj\" nor \"main\"\n");
      }
      if (@blat_files) {
        &print_full($logref,"Found blat alignment files\n");
        $files{$dir}{blat} = \@blat_files;
        $proceed = 1;
      } else {
        &print_full($logref, "No BLAT alignments found\n")
      }
    }

    if ($proceed) {
      #Find the source contigs
      if ($dir eq 'adj') {
        $files{$dir}{contigs} = "$merge_dir/$lib-adj.fa";
      } elsif ($dir eq 'main') {
        $files{$dir}{contigs} = "$merge_dir/$lib-main.fa";
      } else {
        $files{$dir}{contigs} = "$input_dir/$dir/$lib-contigs.fa";
      }
    }
  }
  &print_full($logref, "\n");
  return %files;
}

# Get the file names used for the gap filter
sub Get_Ctg_Seq_File {
  # get function parameters
  my ($blat_out) = @_;
  # DEBUG # &print_full($logref, "BLAT: $blat_out\n");
  my $ctg_seq_file = $blat_out;
  $ctg_seq_file =~ s/output/input/;
  $ctg_seq_file =~ s/\.psl$/.fa/;
  if (not -e "$ctg_seq_file") {
    die "Cannot find contig sequence file: $ctg_seq_file\n";
  }
  return $ctg_seq_file;
}

# Get the arguments needed to use the gap filter on a BLAT alignment file
sub Get_Gapped_Blat_Args {
  # get function parameters
  my ($blat_args, $ctg_seq_file, $gap_out_file) = @_;
  my $gapped_blat_args = join(" ",
    "$blat_args",
    "--use-gap",
    "--ctg-file $ctg_seq_file",
    "--gap-file $gap_out_file",
    "--gap-filter $min_gap_size",
    "--gap-sim $min_gap_similarity");
  # DEBUG # &print_full($logref, "ARGS: $gapped_blat_args\n");
  return $gapped_blat_args;
}

#Run the code to find anomalous contigs on the alignment files
sub Find_Anomalous_Contigs {
  &print_full($logref, "Looking for anomalous contigs...\n");
  for my $dir (sort keys %files) {
    my @commands = ();

    my $ctg_set = '';
    if ($dir eq 'adj') {
      $ctg_set = 'merge.adj';
    } elsif ($dir eq 'main') {
      $ctg_set = 'merge.main';
    } else {
      $ctg_set = $dir;
    }

    my $python = "/gsc/software/linux-x86_64/python-builder-2.5/bin/python";
    my $svn_dir = "/home/lswanson/my_code/svn";
    my $anomalous_ctg_dir =
      "$svn_dir/trans-abyss/trunk/analysis/anomalous_contig_finder";
    my $fac_script =
      "$anomalous_ctg_dir/find_anomalous_contigs_3.0.py";
    my $fac_args = join(" ",
      "--max-match $max_match_fraction",
      "--min-merge-overlap $min_merge_overlap",
      "--min-ctg-cover $min_ctg_cover_fraction",
      "--num-aligns $num_aligns",
      "--min-identity $min_identity");
    if ($OPT{debug}) {
      $fac_args .= " --debug";
    }
    my $append_args = "$fac_args --append";
    my $out_root = "$outdir/${lib}.anomalous_contig.${ctg_set}";

    if (exists $files{$dir}{blat}) {
      my $fac_out = "${out_root}.blat";
      $files{$dir}{blat_fac_out} = $fac_out;
      my $gap_out_file;
      if ($use_gap_filter) {
        $gap_out_file = "${out_root}.gap";
        $files{$dir}{gap_fac_out} = $gap_out_file;
      }
      if (@{$files{$dir}{blat}} == 1) {
        my $blat_out = $files{$dir}{blat}->[0];
        my $blat_args = $fac_args;
        if ($use_gap_filter) {
          my $ctg_seq_file = &Get_Ctg_Seq_File($blat_out);
          $blat_args =
            &Get_Gapped_Blat_Args($blat_args, $ctg_seq_file, $gap_out_file);
        }
        push @commands,
          "$python $fac_script $blat_out $fac_out $blat_args";
      } else {
        if (-e "$fac_out" && ! defined $OPT{ignore_cmds}) {
          push @commands, "rm $fac_out";
        }
        if ($use_gap_filter && -e "$gap_out_file" && ! defined $OPT{ignore_cmds}) {
          push @commands, "rm $gap_out_file";
        }
        for my $blat_out (@{$files{$dir}{blat}}) {
          my $blat_args = $append_args;
          if ($use_gap_filter) {
            my $ctg_seq_file = &Get_Ctg_Seq_File($blat_out);
            $blat_args =
              &Get_Gapped_Blat_Args($blat_args, $ctg_seq_file, $gap_out_file);
          }
          push @commands,
            "$python $fac_script $blat_out $fac_out $blat_args";
        }
      }
    }

    if (exists $files{$dir}{exon}) {
      my $fac_out = "${out_root}.exon";
      $files{$dir}{exon_fac_out} = $fac_out;
      if (@{$files{$dir}{exon}} == 1) {
        my $exon_out = $files{$dir}{exon}->[0];
        push @commands,
          "$python $fac_script $exon_out $fac_out $fac_args";
      } else {
        if (-e "$fac_out" && ! defined $OPT{ignore_cmds}) {
          push @commands, "rm $fac_out";
        }
        for my $exon_out (@{$files{$dir}{exon}}) {
          push @commands,
            "$python $fac_script $exon_out $fac_out $append_args";
        }
      }
    }

    if (defined $OPT{ignore_cmds}) {
      &print_full($logref, "Ignoring commands...\n");
      open(IGNORE,$OPT{ignore_cmds}) || die "Can't open ignore cmds file $OPT{ignore_cmds} $!\n";
      while (<IGNORE>) { # get a line from the file
        chomp;
        my $ignore_cmd = $_;
        my $cmd_index = 0;
        #&print_full($logref, "IGNORE: $ignore_cmd\n");
        for my $command (@commands) {
          #&print_full($logref, "   CMD: $command\n");
          if ("$ignore_cmd" eq "$command") {
            #&print_full($logref, "-- ignore it --\n");
            splice(@commands, $cmd_index, 1);
            last;
          }
          $cmd_index++;
        }
      }
    }

    if ($OPT{run_acf}) {
      my $total_find_anomalous_contig_start = new Benchmark;
      for my $command (@commands) {
        my $file_name = '';
        if ($command =~ /find_anomalous_contigs/) {
          # extract file name from command
          my @cmd_fields = split(" ",$command);
          #my $num_cmd_fields = @cmd_fields;
          #print "# command fields: $num_cmd_fields\n"; # DEBUG
          my $FAC_ARG_ALIGN_FILE = 2;
          ($file_name) =
            $cmd_fields[$FAC_ARG_ALIGN_FILE] =~ /\/([^\/]*)$/;
          if ("" ne $file_name) {
            &print_full($logref, "Searching alignment file: $file_name\n");
          } else {
            die "Could not get file name from \"$cmd_fields[$FAC_ARG_ALIGN_FILE]\"\n";
          }
          &print_full($logref,"$command\n") if $OPT{debug};
        } else {
          &print_full($logref, "$command\n");
        }
        my $find_anomalous_contig_start = new Benchmark;
        my $output = `$command 2>&1`;
        my $result = $?;
        &print_full($logref, "$output");
        # &print_full($logref, "RESULT: $result\n"); # DEBUG
        if ($command =~ /find_anomalous_contigs/) {
          my ($num_new_gaps) =
            $output =~ /Number of gapped alignments found: (\d+)/;
            #&print_full($logref, "New gaps: \"$num_new_gaps\"\n"); # DEBUG
          $num_gaps_found{$dir} += $num_new_gaps;
        }
        if ($SF_SUCCESS != $result) {
          # check that command was "find_anomalous_contigs"
          # REMINDER: create list of warnings, output separately from log file
          if ($SF_NO_ALIGNMENTS == $result &&
              $command =~ /find_anomalous_contigs/) {
            my $warning = "WARNING: no alignments were found in $file_name.";
            &print_full($logref, "$warning\n");
          } else {
            die "Error while running $fac_script: $result\n";
          }
        }
        unless ($command =~ /^rm/) {
          my $find_anomalous_contig_end = new Benchmark;
          my $find_anomalous_contig_time =
            timediff($find_anomalous_contig_end, $find_anomalous_contig_start);
          my $find_anomalous_contig_time_str =
            timestr($find_anomalous_contig_time);
          &print_full($logref,"Time spent finding anomalous contigs: ");
          &print_full($logref,"$find_anomalous_contig_time_str\n\n");
        }
      }
      my $total_find_anomalous_contig_end = new Benchmark;
      my $total_find_anomalous_contig_time =
        timediff($total_find_anomalous_contig_end, $total_find_anomalous_contig_start);
      my $total_find_anomalous_contig_time_str =
        timestr($total_find_anomalous_contig_time);
      &print_full($logref,"Total time spent finding split alignments: ");
      &print_full($logref,"$total_find_anomalous_contig_time_str\n\n");
    } else {
      &print_full($logref, "Anomalous contig files found\n");
      #&print_full($logref, join("\n\n",@commands));
      #&print_full($logref,"\n\n");
    }
  }
}

#Parse the output of find_anomalous_contigs for generating summary file ->
# This gets called multiple times and serves to group the events by chr pairs
sub Parse_Anomalous_Contigs {
  # get function parameters
  my ($file,$aligner,$dir) = @_;
  &print_full($logref, "Parsing anomalous contigs in $file...\n");
  open(FILE,$file) || die "Can't open anomalous contig file $file $!\n";
  #Store the overlapping pairs here
  while (<FILE>) { # get a line from the file
    chomp;
    # &print_full($logref, "$_\n"); # DEBUG
    my @fields = split;
    next if $fields[$ACF_TARGET] =~ /random/ || $fields[$ACF_TARGET] =~ /hap/;
    my ($target1,$target2) = $fields[$ACF_TARGET] =~ /TARGET:(\S+),(\S+)/;
    my ($chr1,$range1) = $target1 =~ /([0-9XYMT]+):(\d+\-\d+)/;
    my ($chr2,$range2) = $target2 =~ /([0-9XYMT]+):(\d+\-\d+)/;
    $chr1 = 'M' if $chr1 eq 'MT';
    $chr2 = 'M' if $chr2 eq 'MT';
    my ($ctg) = $_ =~ /CTG:(\S+)\(/;
    # &print_full($logref, "TARGET: 1) $chr1 2) $chr2\n"); # DEBUG

    #Mark the contig so we parse the kalign alignments for it later
    $contigs_in_events{$ctg}{found} = 1;

    # REMINDER
    # Get the junction co-ordinates from the anomalous contig
    my ($junction_from, $junction_to) =
      $fields[$ACF_JUNCTION] =~ /JUNCTION:(chr[^-]+)-(chr.+)$/;
    my ($chr_from,$coord_from,$dir_from) =
      $junction_from =~ /chr(\d+|[XYM]|MT):(\d+)\((up|down)\)/;
    my ($chr_to,  $coord_to,  $dir_to) =
      $junction_to   =~ /chr(\d+|[XYM]|MT):(\d+)\((up|down)\)/;
    $chr_from = 'M' if $chr_from eq 'MT';
    $chr_to   = 'M' if $chr_to   eq 'MT';
    # &print_full($logref, "JUNCTION: 1) $chr_from 2) $chr_to\n"); # DEBUG

    # sort the junction co-ordinates
    my ($chr_from_s,$coord_from_s,$dir_from_s);
    my ($chr_to_s,$coord_to_s,$dir_to_s);
    my ($range1_s,$range2_s);
    if (($chr_from  < $chr_to) ||
        ($chr_from == $chr_to && $coord_from  < $coord_to) ||
        ($chr_from == $chr_to && $coord_from == $coord_to &&
         "up" eq $dir_from)) {
      $chr_from_s = $chr_from;
      $coord_from_s = $coord_from;
      $dir_from_s = $dir_from;
      $chr_to_s = $chr_to;
      $coord_to_s = $coord_to;
      $dir_to_s = $dir_to;
      $range1_s = $range1;
      $range2_s = $range2;
    } else {
      $chr_from_s = $chr_to;
      $coord_from_s = $coord_to;
      $dir_from_s = $dir_to;
      $chr_to_s = $chr_from;
      $coord_to_s = $coord_from;
      $dir_to_s = $dir_from;
      $range1_s = $range2;
      $range2_s = $range1;
    }
    my $junction_from_s = "$coord_from_s($dir_from_s)";
    my $junction_to_s   = "$coord_to_s($dir_to_s)";

    #&print_full($logref, "JUNCTION:\n"); # DEBUG
    #&print_full($logref, "    $junction_from\n"); # DEBUG
    #&print_full($logref, "  from ${dir_from_s}stream of chr${chr_from_s}:${coord_from_s}\n"); # DEBUG
    #&print_full($logref, "    $junction_to\n"); # DEBUG
    #&print_full($logref, "    to ${dir_to_s}stream of chr${chr_to_s}:${coord_to_s}\n"); # DEBUG

    #Fix exonerate alignments from highest to lowest
    my ($start1a,$end1a) = &Sort_Range($range1_s);
    my ($start1b,$end1b) = &Sort_Range($range2_s);
    if ($OPT{start}) {
      next unless ($start1a =~ /^$OPT{start}/ || $start1b =~ /^$OPT{start}/);
    }
    my $rangeA_sorted = "$start1a-$end1a";
    my $rangeB_sorted = "$start1b-$end1b";
    my $anomalous_contig = $_;
    # DEBUG # &print_full($logref, "BEFORE: $anomalous_contig\n"); # DEBUG
    # Do not rearrange target co-ordinates, it will disconnect them from
    # the query co-ordinates
    #$anomalous_contig =~ s/$range1_s/$rangeA_sorted/;
    #$anomalous_contig =~ s/$range2_s/$rangeB_sorted/;
    # DEBUG # &print_full($logref, " AFTER: $anomalous_contig\n"); # DEBUG
    my $event_grouping_info = join(" ",
      "$rangeA_sorted",
      "$rangeB_sorted",
      "$junction_from_s",
      "$junction_to_s",
      "$dir",
      "$aligner",
      "$anomalous_contig");
    #&print_full($logref, "GRP:\n  $event_grouping_info\n"); # DEBUG
    # Check for duplicates
    if (exists($target_sets{$chr_from_s}{$chr_to_s}{"$event_grouping_info"})) {
      die "Duplicate event found: $event_grouping_info\n";
    }
    # Store the alignment pairs by chr pairs for easier grouping later
    $target_sets{$chr_from_s}{$chr_to_s}{"$event_grouping_info"} = 1;
  }
}

#Group all the individual alignments
sub Group {
  #Now do the pairwise comparisons on the sorted anomalous contigs
  &print_full($logref,"Grouping events...\n");
  my $group_start = new Benchmark;
  #Buffer to add to junction regions to determine whether to group pairs
  my $group_buffer = 50;
  #Keep a unique key for the groups
  my $group_key = 1;

  #Keep track of which lines have been used to avoid adding lines twice
  my %lines_used = ();

  my $set1a = Set::IntSpan->new();
  my $set1b = Set::IntSpan->new();
  my $set2a = Set::IntSpan->new();
  my $set2b = Set::IntSpan->new();

  for my $chr1 (sort {$a<=>$b} keys %target_sets) {
    for my $chr2 (sort {$a<=>$b} keys %{$target_sets{$chr1}}) {
      my @ranges = sort {my ($arange) = $a =~ /^(\d+)/;
                         my ($brange) = $b =~ /^(\d+)/;
                         $arange<=>$brange}
                   keys %{$target_sets{$chr1}{$chr2}};
      my $ranges_str = join("\n", @ranges);
      &print_full($logref, "$ranges_str\n") if $OPT{debug};

      #&print_full($logref,"chr$chr1 v chr$chr2...\n");
      for (my $i=0;$i<@ranges;$i++) {
        #&print_full($logref, "RANGE: $ranges[$i]\n"); # DEBUG
        my @fields1 = split(" ",$ranges[$i]);
        my ($range1a,$range1b,$junction_from1,$junction_to1,$dir1,$aligner1) =
          @fields1;
        my ($ctg1) = $fields1[$SC_CTG_ID] =~ /CTG:(\S+)\(/;
        $ctg_stats{$dir1}{$ctg1}{count}++;

        #Set the local status for the contig if applicable -> create buffered
        # sets and see if they overlap
        my ($start1a,$end1a) = &Sort_Range($range1a);
        my ($start1b,$end1b) = &Sort_Range($range1b);
        if ($chr1 eq $chr2) {
          my $buffer_start1a = $start1a-$local_buffer;
          my $buffer_start1b = $start1b-$local_buffer;
          my $buffer_end1a = $end1a+$local_buffer;
          my $buffer_end1b = $end1b+$local_buffer;
          my $set1 = Set::IntSpan->new("$buffer_start1a-$buffer_end1a");
          my $set2 = Set::IntSpan->new("$buffer_start1b-$buffer_end1b");
          if ($set1->intersect($set2)->cardinality()) {
            $ctg_stats{$dir1}{$ctg1}{local} = 1;
          }
        }

        #Check we haven't already used the entry in a group
        my $num_blocks1 = @fields1;
        my $line1 = join(" ",@fields1[$SC_CTG_ID..($num_blocks1-1)]);
        #&print_full($logref, "1) $line1\n"); # DEBUG
        if (exists $lines_used{"$aligner1 $line1"}) {
          # DEBUG # &print_full($logref, "  Line already used, skipping\n"); # DEBUG
          next;
        }

        # Start a group if this line has not already been used
        # in some other group
        $groups{$group_key}{lines}{"$line1 $aligner1 $dir1"} = 1;

        #Flag that we have found the aligners and sources within the group
        $groups{$group_key}{aligner}{$aligner1} = 1;
        $groups{$group_key}{source}{$dir1} = 1;

        #Mark the line as used
        $lines_used{"$aligner1 $line1"} = 1;

        # REMINDER: This is what is being changed to sort by target
        #  event regions rather than target regions
        #Create the initial sets
        my ($coord_from1,$dir_from1) =
          $junction_from1 =~ /(\d+)\((up|down)\)/;
        if ("" eq $coord_from1) {
          die "Could not get junction co-ordinates from: $junction_from1\n" .
              "   $line1\n";
        }
        my $buffer_start1a = $coord_from1-$group_buffer;
        my $buffer_end1a   = $coord_from1+$group_buffer;
        my $buffer_range1a = "$buffer_start1a-$buffer_end1a";
        $set1a = Set::IntSpan->new($buffer_range1a);
        my ($coord_to1,$dir_to1) =
          $junction_to1 =~ /(\d+)\((up|down)\)/;
        if ("" eq $coord_to1) {
          die "Could not get junction co-ordinates to: $junction_to1\n";
        }
        my $buffer_start1b = $coord_to1-$group_buffer;
        my $buffer_end1b   = $coord_to1+$group_buffer;
        my $buffer_range1b = "$buffer_start1b-$buffer_end1b";
        $set1b = Set::IntSpan->new($buffer_range1b);
        # DEBUG # &print_full($logref, "  Set1: ${set1a}($dir_from1), ${set1b}($dir_to1)\n"); # DEBUG

        #Now check vs the other candidate pairs
        for (my $j=$i+1;$j<@ranges;$j++) {
          my @fields2 = split(" ",$ranges[$j]);
          my $num_blocks2 = @fields2;
          my $line2 = join(" ",@fields2[$SC_CTG_ID..($num_blocks2-1)]);
          #&print_full($logref, "2) $line2\n"); # DEBUG
          my ($range2a,$range2b,
              $junction_from2,$junction_to2,
              $dir2,$aligner2) =
            @fields2;
          # If the event line has already been used, move on to the next line
          if (exists $lines_used{"$aligner2 $line2"}) {
            # DEBUG # &print_full($logref, "  Line already used, skipping\n"); # DEBUG
            next;
          }

          #Buffer the coordinates when checking overlap
          my ($coord_from2,$dir_from2) =
            $junction_from2 =~ /(\d+)\((up|down)\)/;
          if ("" eq $coord_from2) {
            die "Could not get junction co-ordinates from: $junction_from2\n";
          }
          my $buffer_start2a = $coord_from2-$group_buffer;
          my $buffer_end2a   = $coord_from2+$group_buffer;
          my $buffer_range2a = "$buffer_start2a-$buffer_end2a";
          my ($coord_to2,$dir_to2) =
            $junction_to2 =~ /(\d+)\((up|down)\)/;
          if ("" eq $coord_to2) {
            die "Could not get junction co-ordinates to: $junction_to2\n";
          }
          my $buffer_start2b = $coord_to2-$group_buffer;
          my $buffer_end2b   = $coord_to2+$group_buffer;
          my $buffer_range2b = "$buffer_start2b-$buffer_end2b";

          #Create the pair sets to see if they overlap
          my $set2a = Set::IntSpan->new($buffer_range2a);
          my $set2b = Set::IntSpan->new($buffer_range2b);
          # DEBUG # &print_full($logref, "  Set2: ${set2a}($dir_from2), ${set2b}($dir_to2)\n"); # DEBUG

          #If there is overlap
          # REMINDER: should not have to check both directions, since
          # junction regions were sorted
          if ($set1a->intersect($set2a)->cardinality() &&
              $set1b->intersect($set2b)->cardinality() &&
              $dir_from1 eq $dir_from2 &&
              $dir_to1   eq $dir_to2) {
            # DEBUG # &print_full($logref, "   Sets Overlap, combining events\n"); # DEBUG
            # combine event region sets
            $set1a = $set1a->union($set2a);
            $set1b = $set1b->union($set2b);
            #print " ADD\n";
            #Store the info from the line as a key
            $groups{$group_key}{lines}{"$line2 $aligner2 $dir2"} = 1;

            #Flag whether we've found the aligners and sources within the
            # group
            $groups{$group_key}{aligner}{$aligner2} = 1;
            $groups{$group_key}{source}{$dir2} = 1;

            #Mark the line as used
            $lines_used{"$aligner2 $line2"} = 1;
          } else {
            # DEBUG # print "   NOOVER\n";
          }
        }
        # DEBUG # &print_full($logref, "  --\n");
        #increment the group counter
        $group_key++;
      }
    }
  }
  my $group_end = new Benchmark;
  my $group_time =
    timediff($group_end, $group_start);
  my $group_time_str =
    timestr($group_time);
  &print_full($logref,"Time spent grouping events: ");
  &print_full($logref,"$group_time_str\n\n");
}

#Parse the wtss snp lists from Ryan
sub Parse_RNA_Repeats {
  # select the RNA file to use (assume mouse if not human)
  my $human = $lib =~ /HS/?1:0;
  my $rna_file = '';
  if ($human) {
    $rna_file =
      "/home/mapper/lymphoma/ucsc/hg18/repeat/repeatmasker/RNA.repeat.txt";
  } else {
    $rna_file =
      "/home/mapper/Mouse/ucsc/mm9/repeats/repeatmasker/RNA.repeat.txt";
  }
  &print_full($logref,"Parsing RNA repeat file $rna_file...\n");
  open(RNA,$rna_file) || die "Can't open rna file $rna_file $!\n";

  my %rna_sets = ();

  my $parse_rna_start = new Benchmark;
  while (<RNA>) {
    my ($chr,$start,$end,$rname) = split;
    my $rna_set = Set::IntSpan->new("$start-$end");
    if (!exists $rna_sets{"chr$chr"}) {
      $rna_sets{"chr$chr"} = $rna_set;
    } else {
      $rna_sets{"chr$chr"} = $rna_sets{"chr$chr"}->union($rna_set);
    }
  }

  #Now add this read info to the group -> do this lane by lane to avoid memory
  # problems
  for my $group_key (sort {$a<=>$b} keys %groups) {
    #Get the ctgs for the group
    my %event_ctgs = ();
    for my $line_str (keys %{$groups{$group_key}{lines}}) {
      #Parse the relevant info
      my @fields = split(" ",$line_str);
      my ($chr1,$start1,$end1) =
        $fields[$ACF_TARGET] =~ /TARGET:(chr[0-9XYMT]+):(\d+)-(\d+),/;
      my ($chr2,$start2,$end2) =
        $fields[$ACF_TARGET] =~ /TARGET:[^,]*,(chr[0-9XYMT]+):(\d+)-(\d+)/;
      $chr1 = 'M' if $chr1 eq 'MT';
      $chr2 = 'M' if $chr2 eq 'MT';

      #Check the edges against the RNA elements
      my $set1 = Set::IntSpan->new("$start1-$start1");
      my $set2 = Set::IntSpan->new("$end1-$end1");
      my $set3 = Set::IntSpan->new("$start2-$start2");
      my $set4 = Set::IntSpan->new("$end2-$end2");

      if (!$set1->intersect($rna_sets{$chr1})->empty()) {
        $groups{$group_key}{rna}++;
      }

      if (!$set2->intersect($rna_sets{$chr1})->empty()) {
        $groups{$group_key}{rna}++;
      }

      if (!$set3->intersect($rna_sets{$chr2})->empty()) {
        $groups{$group_key}{rna}++;
      }

      if (!$set4->intersect($rna_sets{$chr2})->empty()) {
        $groups{$group_key}{rna}++;
      }
    }
  }
  my $parse_rna_end = new Benchmark;
  my $parse_rna_time =
    timediff($parse_rna_end, $parse_rna_start);
  my $parse_rna_time_str =
    timestr($parse_rna_time);
  &print_full($logref,"Time spent parsing RNA repeat file: ");
  &print_full($logref,"$parse_rna_time_str\n\n");
}

#Run KAligner on the reads and contigs or just get the kaligner output files
sub Run_KAligner {
  # get function parameters
  my ($fasta_files) = @_;

  my $single_target = "$outdir/${lib}_targets.fa";
  my %kalign_out_files = ();
  #concatenated files so remove old copies
  if ($OPT{run_kaligner}) {
    &print_full($logref,"Aligning reads to contigs...\n");
    system("rm $single_target") if -e $single_target;

    for my $fasta_type (keys %{$fasta_files}) {
      if ($fasta_type ne 'bubble') {
        for my $file (@{$fasta_files->{$fasta_type}}) {
          chomp $file;
          # use grep "." to avoid blank lines in $file
          my $cat_cmd = "cat $file | grep \".\" >> $single_target";
          &print_full($logref,"$cat_cmd\n") if $OPT{debug};
          system($cat_cmd);
        }
      } else {
        #here we need to add the k-value to the fasta file so we
        for my $file (@{$fasta_files->{$fasta_type}}) {
          chomp $file;
          my ($k_value) = $file =~ /(k\d\d)/;
          my $sed_cmd = join(" | ",
            "cat $file",
            "sed -e 's/>/>$k_value:/'",
            "grep \".\" >> $single_target");
          &print_full($logref,"$sed_cmd\n") if $OPT{debug};
          system($sed_cmd);
        }
      }
    }

    # Find the files containing the read sequences
    my %read_seq_file_dict = ();
    my $num_rfs = 0;
    if (defined $OPT{read_file_list}) {
      open(READ_FILE_LIST, '<', "$OPT{read_file_list}") ||
        die "Cannot open read file list: $OPT{read_file_list}\n";
      my $file_count = 1;
      while (<READ_FILE_LIST>) {
        chomp;
        my $read_file_path = $_;
        my ($file_name) = $read_file_path =~ /\/([^\/]+)\.[^\.]+/;
        push @{$read_seq_file_dict{$file_name}{$file_count}}, $read_file_path;
        $file_count += 1;
      }
      $num_rfs = $file_count - 1;
    } else {
      # Get all the read sequence files
      my $read_seq_dir =
        defined $OPT{read_seq_dir}?
          $OPT{read_seq_dir}:
          #"$input_dir/reads_to_contigs";
          "/archive/solexa1_*/analysis*/$lib";
      # DEBUG # print "Read sequence files found at: $read_seq_dir\n";
      my @read_seq_files =
        #`ls ${read_seq_dir}/*/*.fq`;
        `ls ${read_seq_dir}/*/*export_f1b6.txt*`;
      $num_rfs = @read_seq_files;

      for my $read_seq_file (@read_seq_files) {
        chomp $read_seq_file;
        my ($flow,$lane) = $read_seq_file =~ /\/([0-9A-Z]+)_(\d)_/;
        push @{$read_seq_file_dict{$flow}{$lane}}, $read_seq_file;
      }
    }
    &print_full($logref, "Num read files found: $num_rfs\n") if $OPT{debug};

    #KAligner
    my $kalign = join("/",
      "/home/traymond/work/abyss/trunk",
      "_xhost-KAligner-r53272/_inst/bin/KAligner");

    #Iterate over the read sequence files and run kaligner ->
    # cleanup the redundant output for the final file
    for my $flow (sort keys %read_seq_file_dict) {
      for my $lane (sort {$a<=>$b} keys %{$read_seq_file_dict{$flow}}) {
        my $kalign_lane_out = "${outdir}/${flow}.kalign$lane";
        $kalign_out_files{$kalign_lane_out} = 1;
        open(KALIGN_FINAL,">$kalign_lane_out") ||
          die "Can't open $kalign_lane_out $!\n";

        for my $read_seq_file (sort @{$read_seq_file_dict{$flow}{$lane}}) {
          my $remove_temp_file = 0;
          # Check line break style in read seq file
          my $file_output = `file $read_seq_file`;
          #&print_full($logref, "EOL: $file_output\n"); # DEBUG
          if ($file_output =~ /CRLF line terminators/) {
            my $warning = "converting end-of-line character for read " .
                          "sequences file $read_seq_file";
            &print_full($logref, "WARNING: $warning\n");
            my $temp_read_seq_file = "$outdir/__temp_read_seq";
            my $sed_cmd = "sed \"s/\r//\" $read_seq_file > $temp_read_seq_file";
            &print_full($logref, "$sed_cmd\n") if $OPT{debug};
            system($sed_cmd);
            $read_seq_file = $temp_read_seq_file;
            $remove_temp_file = 1;
          }
          my $kalign_original = "$outdir/kalign.tmp";
          my $kalign_cmd = join(" > ",
            "$kalign -m -k$read_length $read_seq_file $single_target",
            "$kalign_original");
          &print_full($logref,"$kalign_cmd\n");
          my $run_kalign_start = new Benchmark;
          # need to shift by eight to get the actual return value
          my $result = system($kalign_cmd) >> 8;
          if (0 != $result) {
            die "Error while running $kalign_cmd: $result\n";
          }
          my $run_kalign_end = new Benchmark;
          my $run_kalign_time =
            timediff($run_kalign_end, $run_kalign_start);
          my $run_kalign_time_str =
            timestr($run_kalign_time);
          &print_full($logref,"Time spent running KAligner: ");
          &print_full($logref,"$run_kalign_time_str\n");
          open(KALIGN_FIRST,"$kalign_original") ||
            warn "Can't open file $kalign_original $!\n";
          my $kalign_count = 0; # DEBUG
          while (<KALIGN_FIRST>) {
            # &print_full($logref, "FIRST: $_") if (0 == $kalign_count); # DEBUG
            if (1) { # DEBUG # /k/) {}
              my @fields = split("\t",$_);
              my $read = shift @fields;
              my %contigs = ();
              for my $field (@fields) {
                chomp $field;
                $contigs{$field} = 1;
              }
              #&print_full($logref, "FINAL: $read\t"); # DEBUG
              #&print_full($logref, join("\t",keys %contigs)); # DEBUG
              #&print_full($logref, "\n\n"); # DEBUG
              print KALIGN_FINAL "$read\t",join("\t",keys %contigs),"\n";
            }
            $kalign_count++; # DEBUG
          }
          close KALIGN_FIRST;
          # DEBUG # &print_full($logref,"rm $kalign_original\n");
          if (-e "$kalign_original") {
            system("rm $kalign_original");
          }
          if ($remove_temp_file) {
            system("rm $read_seq_file");
          }
        }
        close KALIGN_FINAL;
      }
    }
  } else {
    &print_full($logref,"Searching for read to contig alignment files...\n");
    #Load the output files that exist otherwise
    opendir(OUTDIR,$outdir) || die "Can't open out dir $outdir $!\n";
    my @kalign_files = grep {/\.kalign[0-9]/} readdir OUTDIR;
    for my $kalign_file (@kalign_files) {
      $kalign_file = "${outdir}/${kalign_file}";
    }
    %kalign_out_files = map {$_ => 1} @kalign_files;
  }
  system("rm $single_target") if -e $single_target;
  return(\%kalign_out_files);
}

sub Get_Contigs_In_Each_Group {
  my %ctgs_in_each_group = ();
  my $status = "Looping through groups to get contigs and " .
               "event regions in each group...";
  &print_full($logref, "$status\n"); # DEBUG
  for my $group_id (sort {$a<=>$b} keys %groups) {
    #&print_full($logref, "  GROUP: $group_id\n"); # DEBUG
    my %ctgs_in_group = ();
    #&print_full($logref, "  Looping through group lines to get event info\n"); # DEBUG
    for my $ctg_info (keys %{$groups{$group_id}{lines}}) {
      #&print_full($logref, "    LINE: $ctg_info\n");
      #Parse the relevant info
      my @ctg_fields = split(" ",$ctg_info);
      my ($ctg_id) = $ctg_fields[$ACF_CTG_ID] =~ /CTG:(\S+)\(/;
      my ($query1,$query2) =
        $ctg_fields[$ACF_CTG_COORDS] =~ /CONTIG:(\S+),(\S+)/;
      my ($event_type) = $ctg_fields[$ACF_TYPE] =~ /TYPE:(.*)/;
      # REMINDER: this may need to change for events found by the gap filter
      my ($breakpoint_start,$breakpoint_end) =
        &Breakpoint_Range($query1, $query2, $event_type);
      $ctgs_in_group{$ctg_id}{event_coords} =
        "$breakpoint_start-$breakpoint_end";
      $ctgs_in_group{$ctg_id}{full_info} = "$ctg_info";
      #&print_full($logref, "$ctg_id: $ctgs_in_group{$ctg_id}\n");
    }
    %{$ctgs_in_each_group{$group_id}} = %ctgs_in_group;
  }
  return \%ctgs_in_each_group;
}

#Parse the kaligner read data
sub Parse_KAligner {
  # get function parameters
  my ($kalign_file_name, $ctgs_in_each_group) = @_;
  open(KALIGN,$kalign_file_name) || die "Can't open $kalign_file_name $!\n";
  &print_full($logref,"Parsing $kalign_file_name...\n");
  my $parse_kalign_start = new Benchmark;
  my %contigs_read_aligns_to = ();
  #&print_full($logref, "Looping through KAlign file\n"); # DEBUG
  while (<KALIGN>) {
    chomp;
    #&print_full($logref, "  KALIGN: $_\n"); # DEBUG
    # next unless /k/; # Do not do this BECAUSE!! contig id does not always start with "k"
    my @kalign_fields = split("\t",$_);
    my $read = shift @kalign_fields;
    #&print_full($logref, "  READ: $read\n"); # DEBUG

    # determine which reads align to each contig
    #&print_full($logref, "  Looping through KAlign fields\n"); # DEBUG
    for my $contig_str (@kalign_fields) {
      #&print_full($logref, "    CTG: $contig_str\n"); # DEBUG
      # Contig ids do not always start with k, especially during testing
      #my ($ctg,$start) = $contig_str =~ /(k\S+)\s+(\d+)/;
      my ($ctg,$start) = $contig_str =~ /^(\S+)\s+(\d+)/;
      #&print_full($logref, "        \"$ctg\"\n"); # DEBUG
      #If the contig has an anomaly
      if (exists $contigs_in_events{$ctg}) {
        #&print_full($logref, "    Adding $read to $ctg read aligns\n");
        my $end = $start+$read_length;
        push @{$contigs_in_events{$ctg}{$kalign_file_name}{read_aligns}},
          "$start-$end $read";

        #Now add all the read mappings to check whether it's unique to
        # the group later on
        #&print_full($logref, "    Looping through the rest of the KAlign fields\n"); # DEBUG
        #for my $contig_str2 (@kalign_fields) {
          # Contig ids do not always start with k, especially during testing
          #my ($ctg2) = $contig_str2 =~ /(k\S+)\s+/;
          #my ($ctg2) = $contig_str2 =~ /(\S+)\s+/;
          #$contigs_read_aligns_to{$read}{$ctg2}++;
        #}
        #last;
      } else { # DEBUG
        #&print_full($logref, "    No $ctg in contigs_in_events\n"); # DEBUG
      } # DEBUG
      # keep track of all the contigs to which the current read aligns
      $contigs_read_aligns_to{$read}{$ctg} = 1;
    }
  }
  close KALIGN;
  &print_full($logref,"Checking read support for events...\n");

  #Now add this read info to the group -> do this lane by lane to avoid memory
  # problems
  #&print_full($logref, "Looping through groups\n"); # DEBUG
  for my $group_key (sort {$a<=>$b} keys %groups) {
    #&print_full($logref, "  GROUP: $group_key\n"); # DEBUG
    # Determine all the contigs that are in the current group
    my %ctgs_in_group = %{$ctgs_in_each_group->{$group_key}};
    #&print_full($logref, "  Looping through group lines to get event info\n"); # DEBUG
    #for my $ctg_info (keys %{$groups{$group_key}{lines}}) {
    #&print_full($logref, "    LINE: $ctg_info\n");
    ##Parse the relevant info
    #my @fields = split(" ",$ctg_info);
    #my ($ctg_id) = $fields[$ACF_CTG_ID] =~ /CTG:(\S+)\(/;
    #my ($query1,$query2) = $fields[$ACF_CTG_COORDS] =~ /CONTIG:(\S+),(\S+)/;
    #my ($breakpoint_start,$breakpoint_end) =
    #&Breakpoint_Range($query1,$query2);
    #$ctgs_in_group{$ctg_id}{event_coords} =
    #"$breakpoint_start-$breakpoint_end";
    #$ctgs_in_group{$ctg_id}{full_info} = "$ctg_info";
    ##&print_full($logref, "$ctg_id: $ctgs_in_group{$ctg_id}\n");
    #}

    #now add the read support info to the group
    #&print_full($logref, "  Looping through event contigs...\n"); # DEBUG
    for my $event_contig (keys %ctgs_in_group) {
      my %good_reads = ();
      #&print_full($logref, "    EVENT CTG: $event_contig\n"); # DEBUG
      if (exists $contigs_in_events{$event_contig} &&
          exists $contigs_in_events{$event_contig}
                                   {$kalign_file_name} &&
          exists $contigs_in_events{$event_contig}
                                   {$kalign_file_name}
                                   {read_aligns}) {
        #&print_full($logref, "EVENT CTG: $event_contig\n"); # DEBUG
        my ($event_start,$event_end) =
          $ctgs_in_group{$event_contig}{event_coords} =~ /(\d+)-(\d+)/;
        #&print_full($logref, "    Looping through reads aligned to current contig\n"); # DEBUG
        for my $read_align
          (@{$contigs_in_events{$event_contig}
                               {$kalign_file_name}
                               {read_aligns}}) {
          #&print_full($logref, "      READ: $read_align\n"); # DEBUG
          my ($read_start,$read_end,$read) =
            $read_align =~ /(\d+)-(\d+)\s+(\S+)/;

          #&print_full($logref, "  $read_start $read_end\t$read\n"); # DEBUG
          #&print_full($logref, "  $event_start $event_end\n---\n"); # DEBUG
          # Check that the read coords covers the event region - if the event
          # region is shorter than the read length, or that the reads falls
          # within the event region - if the event region is longer than the
          # read length
          # REMINDER: this may need to change for events found by
          # the gap filter
          if (($read_start < $event_start && $read_end > $event_end) ||
              ($read_start > $event_start && $read_end < $event_end)) {
            #&print_full($logref, "      Read overlaps event!\n"); # DEBUG
            #Now check that at least 50% of the matches for the read are in
            # the group -> otherwise it isn't unique enough and we can't use
            # it as evidence
            my $num_contigs_in_group = 0;
            #&print_full($logref, "      Looping through contigs read aligns to...\n"); # DEBUG
            for my $read_ctg (keys %{$contigs_read_aligns_to{$read}}) {
              #&print_full($logref, "        EVENT CTG: $event_contig\n"); # DEBUG
              #&print_full($logref, "         READ CTG: $read_ctg\n"); # DEBUG
              if (exists $ctgs_in_group{$read_ctg}) {
                # increase the number of contigs from the current group that
                # the current read aligns to
                $num_contigs_in_group++;
              }
            }
            my $total_num_contigs = keys %{$contigs_read_aligns_to{$read}};
            my $read_fraction = $num_contigs_in_group/$total_num_contigs;
            #&print_full($logref, "Contigs in group: $num_contigs_in_group\n");
            #&print_full($logref, "All Contigs: $total_num_contigs\n");
            #&print_full($logref, "Fraction: $read_fraction\n");
            if ($read_fraction > 0.5) {
              $good_reads{$read} = 1;
              my $msg = join(" ",
                "PASS READ",
                "$num_contigs_in_group/$total_num_contigs",
                "$read($read_start-$read_end)",
                "$event_contig($event_start-$event_end)",
                $ctgs_in_group{$event_contig}{full_info});
              #&print_full($logref, "$msg\n"); # DEBUG
            }
          }
        }
      }
      # add reads for current contig in current group
      # convert array to scalar to count number of items in the array
      my $num_good_reads = keys %good_reads;
      $groups{$group_key}{$event_contig}{read_count} += $num_good_reads;
      if ($OPT{debug}) {
        &print_full($logref, "Add $num_good_reads reads to ");
        &print_full($logref, "g$group_key) $event_contig\n");
        &print_full($logref, "  Current read support: ");
        &print_full($logref,
          "$groups{$group_key}{$event_contig}{read_count}\n");
      }
    }
    #my @good_read_keys = keys %good_reads; # DEBUG
    #print "Good reads: @good_read_keys\n"; # DEBUG
    #for my $read_key (keys %good_reads) { # DEBUG
      #print "  $read_key\n"; # DEBUG
    #} # DEBUG

    # convert array to scalar to count number of items in the array
    #my $num_good_reads = keys %good_reads;
    #&print_full($logref, "Add $num_good_reads reads to $group_key\n"); # DEBUG
    #$groups{$group_key}{read_count} += $num_good_reads;
    #&print_full($logref,
    #"  Current read support: $groups{$group_key}{read_count}\n"); # DEBUG
  }
  my $parse_kalign_end = new Benchmark;
  my $parse_kalign_time =
    timediff($parse_kalign_end, $parse_kalign_start);
  my $parse_kalign_time_str =
    timestr($parse_kalign_time);
  &print_full($logref,"Time spent checking read support: ");
  &print_full($logref,"$parse_kalign_time_str\n");
}

# Summarize the results, this is where the filters are checked and the
# pass/fail status is assigned to each event
sub Summary {
  # get function parameters
  my ($ctgs_in_each_group) = @_;

  my $num_events = keys %groups;
  &print_full($logref, "$num_events potential events found.\n");
  &print_full($logref, "Applying filters...\n");
  #now generate the summary
  my $anomalous_contigs_final_out = "$outdir/${lib}.anomalous_contig.data";
  open(FINAL, ">$anomalous_contigs_final_out") ||
    die "Can't open final output file $anomalous_contigs_final_out $!\n";
  my $anomalous_contigs_pass_out = "$outdir/${lib}.anomalous_contig.pass";
  open(PASS, ">$anomalous_contigs_pass_out") ||
    die "Can't open pass output file $anomalous_contigs_pass_out $!\n";

  #number used for report
  my $group_num = 1;
  my $num_passing_events = 0;
  for my $group_key (sort {$a<=>$b} keys %groups) {
    #record all the reasons the event may fail
    my %fail_reasons = ();

    #Set this to fail by default until we find a group member that passes
    if ($OPT{ctg_report_min}) {
      $fail_reasons{'CTG_REPORT_MIN'} = 1;
    }

    #Set this to fail by default until we find a group member that passes
    if ($OPT{min_read_coverage}) {
      $fail_reasons{'READ_COVERAGE'} = 1;
    }

    #&print_full($logref,
    #"Read support for $group_key: $read_support\n"); # DEBUG

    #Store the k_values for nosingle_k flag
    my %k_values = ();

    #Generate sets for the group members
    my %sets = ();

    #Represents the local status of the group -> a single local contig changes
    # the group status to local
    my $group_local = 'N';

    #Store the group lines as we can't report it until we have the group stats
    my @group_lines = ();
    my %grp_ctgs = ();
    my %group_types = ();
    my $member_count = 'a';
    # iterate over the group's lines (each line represents
    # one anomalous contig alignment pair)
    for my $line_str (keys %{$groups{$group_key}{lines}}) {
      # DEBUG # &print_full($logref, "$line_str\n"); # DEBUG
      #Parse the relevant info
      my @fields = split(" ",$line_str);
      my ($ctg_str) = $fields[$ACF_CTG_ID] =~ /CTG:(\S+)/;
      my ($ctg_id) = $fields[$ACF_CTG_ID] =~ /CTG:(\S+)\(/;
      my ($event_type) = $fields[$ACF_TYPE] =~ /TYPE:(\S+)/;
      my ($target1,$target2) = $fields[$ACF_TARGET] =~ /TARGET:(\S+),(\S+)/;
      my ($chr1,$tstart1,$tend1) = $target1 =~ /([0-9XYMT]+):(\d+)-(\d+)/;
      my ($chr2,$tstart2,$tend2) = $target2 =~ /([0-9XYMT]+):(\d+)-(\d+)/;
      $chr1 = 'M' if $chr1 eq 'MT';
      $chr2 = 'M' if $chr2 eq 'MT';
      my ($query1,$query2) = $fields[$ACF_CTG_COORDS] =~ /CONTIG:(\S+),(\S+)/;
      my $junction_info = $fields[$ACF_JUNCTION];
      my ($aligner,$dir) = ($fields[-2],$fields[-1]);

      # add the event type info for the group
      $group_types{$event_type} = 1;

      my ($start_align) =
        $fields[$ACF_META_DATA] =~ /^(\S+),I1:/;
      #&print_full($logref, "Start align: \"$start_align\"\n"); # DEBUG
      if ("" eq "$start_align") {
        #&print_full($logref, "Use blocks: $fields[$ACF_BLOCKS]\n"); # DEBUG
        ($start_align) = $fields[$ACF_BLOCKS] =~ /(BLOCK:\S+)/;
      }
      if ("" eq "$start_align") {
        die "Could not get meta data from $fields[$ACF_META_DATA]\n";
      }
      my ($i1,$i2,$af1,$af2) =
        $fields[$ACF_META_DATA] =~ /I1:(.*),I2:(.*),AF1:(.*),AF2:(.*)/;

      #Record the k value
      my $k_value = '';
      if ($dir =~ /(k\d\d)/) {
        $k_value = $1;
      } else {
        ($k_value) = $fields[$ACF_CTG_ID] =~ /CTG:(k\d\d)/;
      }
      if ('' == $k_value) {
        $k_value = "default_k";
      }

      #push @{$k_values{$k_value}},$ctg_id;
      $k_values{$k_value}{$ctg_id}++;

      # Store the ids of the contigs in the current group
      $grp_ctgs{$ctg_id} = $dir;
      #Store the contig for generating the fasta later
      #$fasta_data{$dir}{ctgs}{$ctg_id}++;

      #Check if it's a local alignment scenario
      my $neighbors = '';
      if ($chr1 eq $chr2 && abs($tstart1-$tstart2) < $local_buffer) {
        $neighbors = 'neighbor';
      } elsif ($chr1 eq $chr2) {
        $neighbors = 'samechr';
      } else {
        $neighbors = 'diffchr';
      }

      # Should not have to do this, as it was done when the anomalous contig
      # object was initially created
      # order the blocks by query start
      #my ($query1_s, $query2_s, $target1_s, $target2_s);
      #my ($af1_s, $af2_s, $i1_s, $i2_s);
      #&print_full($logref, "qs1: $qstart1, qs2: $qstart2\n"); # DEBUG
      #if ($qstart1 < $qstart2) {
      #$query1_s = $query1;
      #$query2_s = $query2;
      #$target1_s = $target1;
      #$target2_s = $target2;
      #$af1_s = $af1;
      #$af2_s = $af2;
      #$i1_s = $i1;
      #$i2_s = $i2;
      #} else {
      #$query1_s = $query2;
      #$query2_s = $query1;
      #$target1_s = $target2;
      #$target2_s = $target1;
      #$af1_s = $af2;
      #$af2_s = $af1;
      #$i1_s = $i2;
      #$i2_s = $i1;
      #}
      #my $event_coords = join(" ",
      #"ctg:$query1_s=$target1_s($af1_s,$i1_s)",
      #"ctg:$query2_s=$target2_s($af2_s,$i2_s)");

      # get event co-ordinates
      my ($qstart1,$qend1) = $query1 =~ /(\d+)-(\d+)/;
      my ($qstart2,$qend2) = $query2 =~ /(\d+)-(\d+)/;
      my $qspan1 = abs($qstart1 - $qend1) + 1;
      my $qspan2 = abs($qstart2 - $qend2) + 1;
      my $tspan1 = abs($tstart1 - $tend1) + 1;
      my $tspan2 = abs($tstart2 - $tend2) + 1;
      my $event_coords = join(" ",
        "ctg:$query1(${qspan1}bp)=$target1(${tspan1}bp);AF:$af1,PID:$i1",
        "ctg:$query2(${qspan2}bp)=$target2(${tspan2}bp);AF:$af2,PID:$i2");

      #format the target string
      my $target1_ordered = my $target2_ordered = '';
      if ($tstart1>$tend1) {
        $target1_ordered = "chr$chr1:$tend1-$tstart1";
      } else {
        $target1_ordered = "chr$chr1:$tstart1-$tend1";
      }
      if ($tstart2>$tend2) {
        $target2_ordered = "chr$chr2:$tend2-$tstart2";
      } else {
        $target2_ordered = "chr$chr2:$tstart2-$tend2";
      }
      my $set_range1 = my $set_range2 = '';
      if (($chr1 <  $chr2) ||
          ($chr1 == $chr2 && $tstart1 < $tstart2)) {
        ($set_range1) = $target1_ordered =~ /(\d+\-\d+)/;
        ($set_range2) = $target2_ordered =~ /(\d+\-\d+)/;
        if (exists $sets{1}{$chr1}) {
          $sets{1}{$chr1} =
            $sets{1}{$chr1}->union(Set::IntSpan->new($set_range1));
        } else {
          $sets{1}{$chr1} =
            Set::IntSpan->new($set_range1);
        }
        if (exists $sets{2}{$chr2}) {
          $sets{2}{$chr2} =
            $sets{2}{$chr2}->union(Set::IntSpan->new($set_range2));
        } else {
          $sets{2}{$chr2} =
            Set::IntSpan->new($set_range2);
        }
      } else { # if (($chr1 >  $chr2) ||
               #     ($chr1 == $chr2 && $tstart1 >= $tstart2))
        #Here we flip the order
        ($set_range1) = $target2_ordered =~ /(\d+\-\d+)/;
        ($set_range2) = $target1_ordered =~ /(\d+\-\d+)/;
        if (exists $sets{1}{$chr2}) {
          $sets{1}{$chr2} =
            $sets{1}{$chr2}->union(Set::IntSpan->new($set_range1));
        } else {
          $sets{1}{$chr2} =
            Set::IntSpan->new($set_range1);
        }
        if (exists $sets{2}{$chr1}) {
          $sets{2}{$chr1} =
            $sets{2}{$chr1}->union(Set::IntSpan->new($set_range2));
        } else {
          $sets{2}{$chr1} =
            Set::IntSpan->new($set_range2);
        }
      }

      #Check if the contig is ever part of a local alignment
      my $local = 'N';
      #Set the local group status if it's local
      if ($ctg_stats{$dir}{$ctg_id}{local}) {
        $group_local = 'Y';
        $local = 'Y';
        $fail_reasons{'LOCAL'}++ if $OPT{nolocal};
      }

      #Number of times the contig is reported in an event
      my $ctg_count = $ctg_stats{$dir}{$ctg_id}{count};

      #If we find an entry that meets the criteria remove the fail status
      if ($OPT{ctg_report_min} && $ctg_count <= $ctg_report_min) {
        delete $fail_reasons{'CTG_REPORT_MIN'};
      }

      my $read_support = "N/A";
      if ($OPT{min_read_coverage}) {
        my %ctgs_in_group = %{$ctgs_in_each_group->{$group_key}};
        my ($event_start,$event_end) =
          $ctgs_in_group{$ctg_id}{event_coords} =~ /(\d+)-(\d+)/;
        my $event_len = abs($event_start - $event_end) + 1;
        $read_support = $groups{$group_key}{$ctg_id}{read_count};
        my $norm_read_supp = $read_support / $event_len;
        $read_support .= "(${event_len}bp:$norm_read_supp)";
        if ($read_support >= $min_read_coverage) {
          delete $fail_reasons{'READ_COVERAGE'};
        }
      }

      push @group_lines, join(" ",
        "$event_type",
        "$event_coords",
        "$dir:$ctg_str",
        "$start_align",
        "ALIGN:$aligner,$neighbors",
        "READ_TO_CTG:$read_support",
        "CTGLOCAL:$local",
        "CTGCANDIDATE:$ctg_count",
        "$junction_info");
    }

    #Now get the group stats
    #Check if we have one or both aligners
    my $aligners =  '';
    if (exists $groups{$group_key}{aligner}{exon} &&
        exists $groups{$group_key}{aligner}{blat}) {
      $aligners = 2;
    } else {
      $aligners = 1;
      $fail_reasons{'SINGLE_ALIGN'} = 1 if $OPT{require_both_aligners};
    }

    #Check if we have RNA repeats
    if ($OPT{noRNA} && exists $groups{$group_key}{rna}) {
      $fail_reasons{'RNA_REPEAT'} = 1;
    }

    #if we don't find it in enough sources
    my $source_count = keys %{$groups{$group_key}{source}};
    if ($min_source_num > $source_count) {
      $fail_reasons{'TOO_FEW_SOURCES'}++;
    }

    my $group_member_count = keys %{$groups{$group_key}{lines}};

    #print Dumper \%sets;

    #Sort the lines based on sorted coordinates
    my $member_count = 'a';
    my @final_lines = ();
    for my $sortline
      (sort {my ($astart1,$astart2) =
               $a =~ /chr[0-9XYMT]+:(\d+)[^ ]* ?[^ ]*chr[0-9XYMT]+:(\d+)/;
             my ($bstart1,$bstart2) =
               $b =~ /chr[0-9XYMT]+:(\d+)[^ ]* ?[^ ]*chr[0-9XYMT]+:(\d+)/;
             $astart1<=>$bstart1 || $astart2<=>$bstart2} @group_lines) {
      push @final_lines,"$group_num$member_count) $sortline";
      $member_count++;
    }

    # if we find only a single unique (i.e. unmerged) contig and
    # a single k_value
    if ($OPT{nosingle_k}) {
      if (keys %k_values == 1) {
        my $fail = 1;
        for my $k_value (keys %k_values) {
          my %ctgs = %{$k_values{$k_value}};
          for my $ctg (keys %ctgs) {
            # if the contig name ends in 'u' it was not involved in any merges
            $fail = 0 if $ctg !~ /u$/;
          }
        }
        if ($fail) {
          $fail_reasons{'SINGLE_K'} = 1;
        }
      }
    }

    my $chr1 = (keys %{$sets{1}})[0]; # they all have the same chr1
    my $start1 = $sets{1}{$chr1}->min();
    my $coord1 = "chr$chr1:".$sets{1}{$chr1}->min()."-".$sets{1}{$chr1}->max();
    my $chr2 = (keys %{$sets{2}})[0]; # they all have the same chr2
    my $start2 = $sets{2}{$chr2}->min();
    my $coord2 = "chr$chr2:".$sets{2}{$chr2}->min()."-".$sets{2}{$chr2}->max();
    my $coord_str = '';
    if ($chr1 eq $chr2) {
      if ($start1 < $start2) {
        $coord_str = "$coord1,$coord2";
      } else {
        $coord_str = "$coord2,$coord1";
      }
    } elsif ($chr1 > $chr2) {
      $coord_str = "$coord2,$coord1";
    } else {
      $coord_str = "$coord1,$coord2";
    }

    if ($OPT{noMito}) {
      if ("M" eq $chr1 || "M" eq $chr2) {
        $fail_reasons{'MITOCHONDRIAL'} = 1;
      }
    }

    my $status_str = '';
    my $pass = 0;
    if (keys %fail_reasons) {
      $status_str = 'FAIL('.join(",",keys %fail_reasons).')';
    } else {
      $pass = 1;
      $status_str = 'PASS';
    }

    # REMINDER: figure out how to get physical coverage so that I can do this!
    # If we want to set candidates with some minimum read pair (aka physical)
    # coverage to pass as long as they are not RNA
    # TEMP # if (defined $OPT{read_pair_trump}) {
    # TEMP # if ($OPT{read_pair_trump} < $read_count &&
    # TEMP # !exists $fail_reasons{'RNA_REPEAT'}) {
    # TEMP # $status_str =~ s/FAIL/PASS/;
    # TEMP # }
    # TEMP # }

    # count the number of contigs appearing in the group
    # DEBUG # &print_full($logref, "Counting contigs for group $group_num\n");
    my $num_contigs = 0;
    for my $k_value (keys %k_values) {
      # DEBUG # &print_full($logref, "  K Value: $k_value\n");
      $num_contigs += keys %{$k_values{$k_value}};
      for my $ctg (keys %{$k_values{$k_value}}) {
        # DEBUG # &print_full($logref, "    Ctg: $ctg\n");
      }
    }
    # DEBUG # &print_full($logref, "----\n");

    my $types = join(",", sort keys %group_types);

    my $group_str = join(" ",
      "GRPNUM:$group_num",
      "TYPE:$types",
      "COORDS:$coord_str",
      "MEMBERS:$group_member_count",
      "ALIGNERS:$aligners",
      "SOURCES:$source_count",
      "CONTIGS:$num_contigs",
      "LOCAL:$group_local",
      # this should be separate for each contig
      # TEMP # "READPAIR:$read_count",
      "STATUS:$status_str\n");

    my $line = $group_str.join("\n",@final_lines)."\n\n";
    print FINAL $line;
    if ($status_str =~ /PASS/) {
      #Store the contig for generating the fasta later
      for my $ctg (keys %grp_ctgs) {
        my $dir = $grp_ctgs{$ctg};
        $fasta_data{$dir}{ctgs}{$ctg} = 1;
      }
      $num_passing_events++;
      print PASS $line;
    }
    $group_num++;
  }
  &print_full($logref, "$num_passing_events events passed all filters.\n");
}

#Determine the estimated breakpoint coord(s) depending on the alignments
sub Breakpoint_Range {
  # get function parameters
  my ($r1, $r2, $event_type) = @_;
  my ($s1, $e1) = $r1 =~ /(\d+)-(\d+)/;
  my ($s2, $e2) = $r2 =~ /(\d+)-(\d+)/;

  my $low_start = my $low_end = my $high_start = my $high_end = '';
  # First we sort the range (do this because events found by the gapfilter
  # will not necessarily have these co-ordinates sorted already
  if ($s1 < $s2) {
    $low_start = $s1;
    $low_end = $e1;
    $high_start = $s2;
    $high_end = $e2;
  } else {
    $low_start = $s2;
    $low_end = $e2;
    $high_start = $s1;
    $high_end = $e1;
  }

  my $overlap_start = my $overlap_end = '';
  if ($low_end <= $high_start) {
    #here it's blunt or there's a gap
    $overlap_start = $low_end - 1;
    $overlap_end = $high_start + 1;
    # if the event was found by the gap filter
    if ($event_type =~ /gap/) {
      &print_full($logref, "Getting breakpoint for gap event\n"); # DEBUG
      # determine the distance between the contig regions
      my $distance = $overlap_end - $overlap_start + 1;
      # if the distance is larger than the read length
      if ($distance >= $read_length) {
        # use the gap2ctg alignment target co-ordinates
        $overlap_start = $s2 - 1;
        $overlap_end = $e2 + 1;
      }
    }
  } else {
    #here they overlap
    $overlap_start = $high_start - 1;
    $overlap_end = $low_end + 1;
  }

  return ($overlap_start,$overlap_end);
}

#Sorts the range for Set::IntSpan input
sub Sort_Range {
  # get function parameters
  my ($range) = @_;
  my ($coord1,$coord2) = $range =~ /(\d+)-(\d+)/;
  if ($coord1<$coord2) {
    return ($coord1,$coord2);
  } else {
    return ($coord2,$coord1);
  }
}

#Get the contigs to create a separate fasta file
sub Fasta {
  #Generate the fasta file
  my $anomalous_contigs_fasta_out = "$outdir/${lib}.anomalous_contig.contigs";
  open(FASTAOUT,">$anomalous_contigs_fasta_out") ||
    die "Can't open $anomalous_contigs_fasta_out $!\n";
  for my $dir (keys %fasta_data) {
    #first create the fasta
    open(FASTAIN,$fasta_data{$dir}{fasta}) ||
      die "Can't open $fasta_data{$dir}{fasta} $!\n";
    my @lines = <FASTAIN>;
    for (my $i=0;$i<@lines;$i++) {
      my $ctg = '';
      my $cov = '';
      if ($lines[$i] =~ /^>/) {
        ($ctg,$cov) = $lines[$i] =~ /^>(\S+)(.*)$/;
      }
      if (exists $fasta_data{$dir}{ctgs}{$ctg}) {
        print FASTAOUT ">$dir:$ctg$cov\n";
        print FASTAOUT "$lines[$i+1]\n";
        $ctg = '';
      }
    }
    close FASTAIN;
  }
  close FASTAOUT;
}

#Print to stdout and a filehandle at the same time
sub print_full {
  my ($handle,$string) = @_;
  print $handle $string;
  print $string;
}
