#! /usr/bin/env python

# import standard modules
from optparse import OptionParser
import sys, os
import subprocess

# use Readman's wtss modules
sys.path.append("/home/rchiu/work/wtss")

# import custom modules
from wtss.utilities.align_parsers import blat, exonerate
# use blat.parse(file, filters, noblocks=False noline=False)
# use exonerate.parse(file, filters)
# filters: a dictionary with the following keys relevant
#   bestn:  for each query, return all alignments with rank <= bestn
#   unique: for each query with a single best alignment, return that alignment
#   count:  for each query, return the first count alignments
#   qlen:   for each query, return all alignments with query length >= qlen
#   match:  for each query, return all alignments with >= match% of query bases
#           matching
#   identity: for each query, return all alignments with percent identity >=
#             identity
#   ungapped: for each query, return all alignments with only a single block
# Use custom intspan library
from wtss.utilities import intspan
from anomalous_contig_3 import AnomalousContigCls, fix_align

# CONSTANTS
ES_SUCCESS = 0
ES_SPLIT_FINDER_ERR = 1
ES_OPTIONS_ERR = 2
ES_NO_ALIGNMENTS = 3
GAP_FILTER_FAIL = 0
GAP_FILTER_PASS = 1
GAP_BLATTER_FAIL = 0
GAP_BLATTER_MIN_SUCCESS = 1
GAP_BLATTER_MAX_SUCCESS = 100
SVN_DIR = "/home/lswanson/my_code/svn"
ANOMALOUS_CTG_DIR = \
  os.path.join(SVN_DIR, "trans-abyss/trunk/analysis/anomalous_contig_finder")

# Class to identify interesting split alignments (chimeric RNA, exon
# duplication, local inversions)
class SplitFinder:
  def __init__(self, params, debug=False):
    self.params = params
    self.anomalous_contigs = []
    self.num_gapped_aligns = 0
    self.more_than_99 = False
    self.debug = debug

    if debug:
      err_msg(self.params)
    # end if
  # end def

  # parse the alignment file and call the function to find the split
  # alignments
  def find(self, input_file_name):
    # parse alignment file
    aligns = self.parse_alignment_file(input_file_name)
    err_msg("Done parsing alignments")
    if (None == aligns or 0 == len(aligns)):
      msg = "No alignments were found."
      raise NoAlignmentsError(msg)
    # end if
    # identify anomalous contigs
    self.get_anomalous_contigs(aligns)
    err_msg("Done identifying anomalous contigs")
    err_msg("Number of alignments: %d" % len(aligns))
    err_msg("Number of split alignments found: %d" % len(self.anomalous_contigs))
    if (self.params['use_gap_filter']):
      msg = "Number of gapped alignments found: "
      if (self.more_than_99):
        msg += "at least "
      # end if
      msg += "%d" % self.num_gapped_aligns
      err_msg(msg)
    # end if
  # end def

  def parse_alignment_file(self, input_file_name):
    aligns = None
    # filters to use when parsing
    filters = \
      {'count':self.params['num_aligns'],
       'identity':self.params['min_identity']}
    # get the input file name extension to determine the tool
    # used for the alignment
    ext = os.path.splitext(input_file_name)[1]
    # parse the alignment file
    if(ext == ".psl"):
      aligns = \
        blat.parse(input_file_name, filters)
    elif(ext == ".exon"):
      aligns = \
        exonerate.parse(input_file_name, filters)
      if (self.params['use_gap_filter']):
        msg = "Cannot use gap filter with exonerate alignments"
        raise SplitFinderError(msg)
      # end if
    else:
      msg  = "Unrecognized input file extension: \"%s\". " % ext
      msg += "Should be either \".exon\" or \".psl\""
      raise SplitFinderError(msg)
    # end if
    return aligns
  # end def

  # identify alignment pairs that look like they could represent
  # interesting split alignments
  def get_anomalous_contigs(self, aligns):
    # TEMP # if self.debug:
      # TEMP # print_aligns(aligns)
    # end if

    # open the contig sequences file if using the gap filter
    if (self.params['use_gap_filter']):
      fail_msg = "Cannot open contig sequence file"
      ctg_seq_file = openfile(self.params['ctg_seq_file'],"r",fail_msg)
    # end if

    # iterate over the alignments, grouping them by query (i.e. contig)
    contig_align_index = 0
    while (contig_align_index < len(aligns)):
      best_aligns = None
      curr_contig = aligns[contig_align_index].query
      #err_msg("Getting best alignments for %s" % curr_contig) # DEBUG
      #err_msg("Contig align index: %i" % contig_align_index) # DEBUG
      # iterate over the alignments to the current contig
      for curr_align_index in range(contig_align_index, len(aligns)):
        #err_msg("Curr Align Index: %i" % curr_align_index) # DEBUG
        #err_msg("Curr Align Ctg: %s" % aligns[curr_align_index].query) # DEBUG
        # stop looking when we are onto the next contig
        if (aligns[curr_align_index].query != curr_contig):
          #err_msg("Moving on to next contig")
          break
        # end if

        curr_align = fix_align(aligns[curr_align_index])

        #err_msg("%i) %s %s" %
          #(curr_align_index, curr_align.query,curr_align.score)) # DEBUG

        # if the alignment method is blat
        if (self.params['use_gap_filter']):
          #err_msg("Running gap filter") # DEBUG
          if (None == curr_align.psl_str):
            msg = "Cannot use gap filter on alignment without psl string"
            raise SplitFinderError(msg)
          # end if
          psl_str = curr_align.psl_str
          # get the contig info from the contig sequence file
          if (curr_align_index == contig_align_index):
            ctg_info = ctg_seq_file.next()
            if (">" != ctg_info[0]):
              msg = "Error reading contig sequence file: %s" % ctg_info
              raise SplitFinderError(msg)
            # end if
            ctg_id = ctg_info[1:].split(None,1)[0]
            while (ctg_id != curr_align.query):
              msg = "Warning: contig %s not present in alignment file" % ctg_id
              err_msg(msg)
              # skip the sequence
              ctg_seq_file.next()
              # get the next contig id
              try:
                ctg_info = ctg_seq_file.next()
              except StopIteration, e:
                msg  = "Error: end of contig sequence file reached before "
                msg += "end of alignment file."
                raise SplitFinderError(msg)
              # end try
              if (">" != ctg_info[0]):
                msg = "Error reading contig sequence file: %s" % ctg_info
                raise SplitFinderError(msg)
              # end if
              ctg_id = ctg_info[1:].split(None,1)[0]
            # end while
          # end if
          # DEBUG # err_msg("Align: %s" % curr_align.query)
          # DEBUG # err_msg(" Ctgs: %s" % ctg_id)
          if (ctg_id != curr_align.query):
            msg = "Error: contig sequence not in same order as alignments."
            msg += " From file: %s." % ctg_id
            msg += " From alignment: %s." % curr_align.query
            raise SplitFinderError(msg)
          # end if
          if (curr_align_index == contig_align_index):
            ctg_seq = ctg_seq_file.next()
            ctg_seq = ctg_seq.rstrip()
          # end if
          #err_msg("SEQ: \"%s\"" % ctg_seq) # DEBUG
          # DEBUG # err_msg("Blat alignment:")
          #err_msg("  %s" % psl_str) # DEBUG
          # check whether it is a gapped alignment
          gap_filter = os.path.join(ANOMALOUS_CTG_DIR, "gapfilter")
          gap_filter_cmd = "%s \"%s\" %i" % \
            (gap_filter, psl_str, self.params['gap_filter_threshold'])
          filter_return = subprocess.call(gap_filter_cmd, shell=True)
          if (self.debug):
            err_msg("Gap filter returned: %i" % filter_return) # DEBUG
          # end debug
          # if an interesting gap was found
          if (GAP_FILTER_PASS == filter_return):
            if (self.debug):
              err_msg("--alignment passed gap filter") # DEBUG
            # end if
            # REMINDER: run next gap-checking step
            # REMINDER: if the next gap-checking step returns success
            #           do not process the current alignment group any further
            gap_blat = os.path.join(ANOMALOUS_CTG_DIR, "gapblatterDir")
            gap_blat_cmd = "%s \"%s\" %s %s %i %.2f" % \
              (gap_blat, psl_str, ctg_seq, self.params['gap_out_file'],
               self.params['gap_filter_threshold'],
               self.params['min_gap_similarity'])
            #err_msg("Gap Blat Command:") # DEBUG
            #err_msg("    %s" % gap_blat_cmd) # DEBUG
            if (self.params['arg_max'] < len(gap_blat_cmd)):
              msg = "Cannot use gap blat on contig %s: sequence is too long"
              raise SplitFinderError(msg)
            # end if
            # DEBUG # err_msg("Checking gapped alignment with blat...")
            blatter_return = subprocess.call(gap_blat_cmd, shell=True)
            if (self.debug):
              err_msg("Gap blat returned: %i" % blatter_return) # DEBUG
            # end if
            if (GAP_BLATTER_MIN_SUCCESS <= blatter_return and
                GAP_BLATTER_MAX_SUCCESS >= blatter_return):
              if (GAP_BLATTER_MAX_SUCCESS == blatter_return):
                self.more_than_99 = True
              # end if
              self.num_gapped_aligns += blatter_return
              best_aligns = None
              if (self.debug):
                err_msg("Gapped alignment found for %s" % ctg_id)
              # end if
              break
            elif (GAP_BLATTER_FAIL != blatter_return):
              msg = "Error running gap blat: %i" % blatter_return
              raise SplitFinderError(msg)
            # end if
          elif (GAP_FILTER_FAIL != filter_return):
            msg = "Error running gap filter: %i" % filter_return
            raise SplitFinderError(msg)
          # DEBUG # else:
            # DEBUG # err_msg("--alignment failed gap filter")
          # end if
        # end if

        # if the current alignment is long enough to represent
        # a single alignment
        # (do not use the match_len member of the alignment class)
        match_len = float(curr_align.qend) - \
                    float(curr_align.qstart) + 1
        match_fraction = match_len / float(curr_align.query_len)
        if (match_fraction >= self.params['longest_single_align']):
          if (self.debug):
            err_msg("Single Alignment Found for: %s" %
                    curr_align.query)
          # end if
          best_aligns = None
          break
        # end if
        # if the best alignments list is empty
        if (None == best_aligns):
          # TEMP # if (self.debug):
            # TEMP # err_msg("Initializing best_aligns...")
          # end if
          # initialize the best alignments list
          best_aligns = [curr_align]
        # if the best alignments list is not empty,
        # consider the current alignment
        else:
          add_align_to_bests = True
          align_in_bests = False
          if (self.debug):
            err_msg("Curr: ", no_newline=True)
            print_short_align(curr_align)
          # end if
          for index in range(len(best_aligns)):
            # get the overlap fraction between the current alignment and
            # the current best alignment
            if (self.debug):
              err_msg("Best: ", no_newline=True)
              print_short_align(best_aligns[index])
            # end if
            (overlap, overlap_fraction) = \
              self.calc_overlap(curr_align, best_aligns[index])
            if (self.debug):
              err_msg("Overlap: %i Fraction: %f" % (overlap, overlap_fraction))
            # end if
            # if the overlap fraction is great enough
            if (params['min_merge_overlap'] <= overlap_fraction):
              # note that the current alignment overlaps a best alignment
              add_align_to_bests = False
              # if the current alignment score is better than the
              # score of the current best alignment
              if (curr_align.score > best_aligns[index].score):
                if (align_in_bests):
                  if (self.debug):
                    msg = "Removing alignment to: %s" % \
                          best_aligns[index].target
                    err_msg(msg)
                  # end if
                  del best_aligns[index]
                else:
                  if (self.debug):
                    msg = "Replacing alignment to: %s" % \
                          best_aligns[index].target
                    err_msg(msg)
                  # end if
                  # replace the current best alignment with
                  # the current alignment
                  best_aligns[index] = curr_align
                # end if
                align_in_bests = True
              # if the current alignment score is equal to the
              # score of the current best alignment
              elif (curr_align.score == best_aligns[index].score):
                # add the alignment to the list of bests anyway
                add_align_to_bests = True
              # end if
            # end if
          # end for
          if (add_align_to_bests):
            if (self.debug):
              err_msg("Adding alignment to: %s" % curr_align.target)
            # end if
            best_aligns.append(curr_align)
            align_in_bests = True
          elif (self.debug and not align_in_bests):
            err_msg("Ignoring alignment to: %s" % curr_align.target)
          # end if
        # end if
        # TEMP # err_msg("Num Best: %i" % len(best_aligns))
        # TEMP # err_msg("-----")
      # end for
      # DEBUG # if (curr_align.query == curr_contig):
        # DEBUG # err_msg("Skipping alignments to: %s (single alignment found)" % curr_contig)
      # end if
      while (len(aligns) > curr_align_index and
             aligns[curr_align_index].query == curr_contig):
        #err_msg("Skipping alignments for %s" % curr_contig) # DEBUG
        curr_align_index += 1
      # end while
      num_aligns_to_contig = curr_align_index - contig_align_index
      if (self.debug):
        err_msg("Ctg: %s, Num Aligns: %d" %
                (curr_contig, num_aligns_to_contig))
      # end if
      contig_align_index += num_aligns_to_contig
      if (None == best_aligns):
        if (self.debug):
          err_msg("No best alignments found")
        # end if
        continue
      elif (self.debug):
        err_msg("Found %i best aligns:" % len(best_aligns))
        print_aligns(best_aligns)
      # end if

      # iterate over the best alignments pairs for the current contig
      for i in range(len(best_aligns)):
        for j in range(i+1, len(best_aligns)):
          # calculate how much the query regions of the two alignments overlap
          (overlap, overlap_fraction) = \
            self.calc_overlap(best_aligns[i], best_aligns[j])

          # if the alignments should have been merged, but were not (because
          # they had equal scores)
          if (params['min_merge_overlap'] <= overlap_fraction):
            if (self.debug):
              msg = "Not comparing equivalent alignments to: %s and %s" % \
                    (best_aligns[i].target, best_aligns[j].target)
              err_msg(msg)
            # end if
            # do not consider them as a paired alignment
            continue
          # end if

          # calculate the fraction of the total contig that
          # the two alignments represent
          (contig_cover, contig_cover_fraction) = \
            self.calc_ctg_cover_by_aligns(best_aligns[i],
                                          best_aligns[j],
                                          overlap)
          if self.debug:
            print_short_align(best_aligns[i])
            print_short_align(best_aligns[j])
            err_msg("Contig Coverage: %d" % contig_cover)
            err_msg("Contig Coverage Fraction: %.2f" % contig_cover_fraction)
          # end if
          # check whether the contig coverage is too low
          if(contig_cover_fraction <= self.params['min_ctg_represented']):
            if self.debug:
              err_msg("Not enough contig coverage, skipping alignment pair")
            # end if
            continue
          # end if

          # if we get to this point, the alignment pair is worth keeping
          anomalous_contig = \
              AnomalousContigCls(best_aligns[i], best_aligns[j])
          anomalous_contig.overlap_query_fraction  = overlap_fraction
          anomalous_contig.contig_cover_fraction = contig_cover_fraction

          # attempt to determine the event type
          anomalous_contig.determine_type()

          # add the anomalous contig to the list
          self.anomalous_contigs.append(anomalous_contig)

        # end for
      # end for
    # end while
    # close the contig sequences file if using the gap filter
    if (self.params['use_gap_filter']):
      ctg_seq_file.close()
    # end if
  # end def

  def calc_overlap(self, align1, align2):
    # convert the range co-ordinates to integers
    span1 = [int(align1.qstart), int(align1.qend)]
    span2 = [int(align2.qstart), int(align2.qend)]
    # sort the range co-ordinates so that the smaller co-ordinate is first
    span1.sort(lambda x,y:x-y)
    span2.sort(lambda x,y:x-y)
    # calculate the overlap using the intspan module
    overlap = intspan.overlap(span1, span2)
    # calculate the smaller alignment
    # (do not use the match_len member of the alignment class)
    match_len1 = span1[1] - span1[0] + 1
    match_len2 = span2[1] - span2[0] + 1
    # TEMP # err_msg("ML1:%i ML2:%i" % (match_len1, match_len2))
    min_match_len = min(match_len1, match_len2)
    overlap_fraction = float(overlap) / float(min_match_len)
    return (overlap, overlap_fraction)
  # end def

  # calculate the amount of the contig that is represented by the combination
  # of the two alignments and the fraction of the total contig length that
  # this represents
  def calc_ctg_cover_by_aligns(self, align1, align2, olap):
    covered_by_align1 =  int(align1.qend) - int(align1.qstart) + 1
    covered_by_align2 =  int(align2.qend) - int(align2.qstart) + 1
    total_covered = covered_by_align1 + covered_by_align2 - olap
    fraction_covered = float(total_covered) / float(align1.query_len)
    return (total_covered, fraction_covered)
  # end def

  #Generate the output file
  def output(self, out_file, append):
    # open the output file in the appropriate mode
    if append:
      mode = "a"
    else:
      mode = "w"
    # end if
    fail_msg = "Cannot open split alignment output file"
    out = openfile(out_file, mode, fail_msg)
    # write the split alignment details to the output file
    for anomalous_contig in self.anomalous_contigs:
      if ("chr" != anomalous_contig.align1.target[0:3]):
        anomalous_contig.align1.target = \
            "chr%s" % anomalous_contig.align1.target
        #err_msg("  Target: %s" % anomalous_contig.align1.target)
        #msg = "Improperly formatted alignment: %s" % anomalous_contig.details()
        #raise SplitFinderError(msg)
      # end if
      if ("chr" != anomalous_contig.align2.target[0:3]):
        anomalous_contig.align2.target = \
            "chr%s" % anomalous_contig.align2.target
      # end if
      out.write(anomalous_contig.details() + "\n")
    # end for
    out.close()
  # end def
# end class

def print_aligns(aligns):
  err_msg("---")
  for i in range(len(aligns)):
    err_msg("%i)" % i, no_newline=True)
    print_align(aligns[i])
  # end for
  err_msg("---")
# end def

def print_align(align):
  err_msg("S:%i ID:%.3f ML:%i QL:%i Q:%s QS:%i QE:%i T:%s TS:%i TE:%i" %
          (int(align.score), float(align.identity),
           int(align.match_len), int(align.query_len),
           align.query,  int(align.qstart), int(align.qend),
           align.target, int(align.tstart), int(align.tend)))
# end def

def print_short_align(align):
  err_msg("S:%i ID:%.1f Q:%s QS:%i QE:%i QL:%i T:%s" %
          (int(align.score), float(align.identity), align.query,
           int(align.qstart), int(align.qend),
           int(align.query_len), align.target))
# end def

def openfile(filename, mode, fail_msg):
  try:
    file = open(filename, mode)
  except IOError, e:
    msg = "%s: %s\n    %s" % (fail_msg, filename, e)
    raise SplitFinderError(msg)
  # end try
  return file
# end def

def SetupOptionsParser():
  description_string = \
    "Filters contig alignments to find contigs resulting in interesting " \
    "split alignments (representing chimeric RNAs, exon duplications, or " \
    "local inversions)."
  usage_string = "%prog ALIGNMENT_FILE SPLIT_OUTPUT_FILE [ OPTIONS ]"
  parser = OptionParser(description=description_string,
                        version="%prog 2.0",
                        usage=usage_string)
  parser.add_option("-m", "--max-match",
                    type='float', dest="longest_single_align",
                    metavar="FRACTION",
                    help="If any single alignment for a contig uses at least "
                         "FRACTION% of the contig length, do not look for "
                         "any split alignments for that contig. "
                         "[default: %default]")
  parser.add_option("-q", "--min-merge-overlap",
                    type='float', dest="min_merge_overlap", metavar="FRACTION",
                    help="If at least FRACTION% of two alignments represent "
                         "the same region of the contig, consider only the "
                         "higher scoring of the two alignments. "
                         "[default: %default]")
  parser.add_option("-c", "--min-ctg-cover",
                    type='float', dest="min_ctg_represented",
                    metavar="FRACTION",
                    help="Only consider split alignments for which the two "
                         "alignments represent at least FRACTION% of the "
                         "total contig. [default:%default]")
  parser.add_option("-n", "--num-aligns",
                    type='int', dest="num_aligns", metavar="N",
                    help="Extract the best N alignments for each contig "
                         "from the alignments file. [default: %default]")
  parser.add_option("-i", "--min-identity",
                    type='float', dest="min_identity", metavar="PID",
                    help="Extract from the alignments file only alignments "
                         "with a percent identity of at least PID%. "
                         "[default: %default]")
  parser.add_option("--no-gap",
                    action="store_false", dest="use_gap_filter",
                    help="Do not run the gap filter. [default]")
  parser.add_option("--use-gap",
                    action="store_true", dest="use_gap_filter",
                    help="Run the gap filter.")
  parser.add_option("--ctg-file",
                    dest="ctg_seq_file", metavar="FILE",
                    help="Read in contig sequences from FILE. This option is "
                         "mandatory with the --use-gap option and ignored "
                         "otherwise.")
  parser.add_option("--gap-file",
                    dest="gap_out_file", metavar="FILE",
                    help="Write gapped alignments to FILE. This option is "
                         "mandatory with the --use-gap option and ignored "
                         "otherwise.")
  parser.add_option("-g", "--gap-filter",
                    type='int', metavar="N",
                    help="When searching gapped alignments, only consider "
                         "alignments with a query gap of at least N bases "
                         "[default: %default]")
  parser.add_option("--gap-sim",
                   type='float', dest="min_gap_similarity", metavar="F",
                   help="When searching gapped alignments, require that the "
                        "gap sequence aligns back to the rest of the contig "
                        "with at least F% of the gap bases matching "
                        "[default: %default]")
  parser.add_option("-d", "--debug",
                    action="store_true", dest="debug",
                    help="Print debug information while the program runs.")
  parser.add_option("-a", "--append",
                    action="store_true", dest="append",
                    help="Append to the output file, rather than "
                         "overwriting it.")
  parser.set_defaults(longest_single_align=0.95,
                      min_merge_overlap=0.85,
                      min_ctg_represented=0.85,
                      num_aligns=10,
                      min_identity=90.0,
                      use_gap_filter=False,
                      gap_filter=20,
                      min_gap_similarity=0.95,
                      debug=False,
                      append=False)
  return parser
# end def

def GetParams(options):
  params = {}
  params['longest_single_align'] = options.longest_single_align
  params['min_merge_overlap'] = options.min_merge_overlap
  params['min_ctg_represented'] = options.min_ctg_represented
  params['num_aligns'] = options.num_aligns
  params['min_identity'] = options.min_identity
  params['use_gap_filter'] = options.use_gap_filter
  params['ctg_seq_file'] = options.ctg_seq_file
  params['gap_out_file'] = options.gap_out_file
  params['gap_filter_threshold'] = options.gap_filter
  params['min_gap_similarity'] = options.min_gap_similarity
  # if using the gap filter, get the argument character limit
  if (params['use_gap_filter']):
    cmd = "getconf ARG_MAX"
    output = \
      subprocess.Popen(["getconf", "ARG_MAX"],
                       stdout=subprocess.PIPE).communicate()[0]
    params['arg_max'] = int(output)
  # end if
  return params
# end def

#### EXCEPTION CLASSES ####
class MyError(Exception):
  def __init__(self, msg):
    self.msg = msg
  def __str__(self):
    return repr(self.msg)
# end class

class SplitFinderError(MyError):
  """Exception raised for errors encountered by the SplitFinder class"""
  pass
# end class

class NoAlignmentsError(SplitFinderError):
  pass
# end class

def err_msg(msg, no_newline=False):
  """Write an error message to stderr (appends a newline)"""
  if (no_newline):
    print >>sys.stderr, msg,
  else:
    print >>sys.stderr, msg
  # end if
# end def

if __name__ == '__main__':
  # get options, args, and params
  parser = SetupOptionsParser()
  (options, args) = parser.parse_args()
  exit_status = ES_SUCCESS
  if (options.use_gap_filter):
    if (None == options.ctg_seq_file):
      msg = "you must specify a file to read contig sequences from when "
      msg += "using the gap filter (--ctg-file option)"
      parser.error(msg)
      exit_status = ES_OPTIONS_ERR
    elif (None == options.gap_out_file):
      msg = "you must specify a file to write gapped alignments to when "
      msg += "using the gap filter (--gap-file option)"
      parser.error(msg)
      exit_status = ES_OPTIONS_ERR
    elif (not options.append and os.path.exists(options.gap_out_file)):
      os.remove(options.gap_out_file)
    # end if
  # end if
  params = GetParams(options)
  # if the right number of args was used
  if(0 == exit_status and 2 == len(args)):
    input_file_name = args[0]
    split_output_file_name = args[1]
    split_finder = SplitFinder(params, debug=options.debug)
    try:
      split_finder.find(input_file_name)
    except NoAlignmentsError, e:
      err_msg(e)
      exit_status = ES_NO_ALIGNMENTS
    except SplitFinderError, e:
      msg = "Error while searching for anomalous contigs: %s" % e
      err_msg(msg)
      exit_status = ES_SPLIT_FINDER_ERR
    # end try
    if(0 == exit_status):
      split_finder.output(split_output_file_name, append=options.append)
      if(options.debug):
        err_msg("Output to: %s" % split_output_file_name)
      # end if
    # end if
  else:
    parser.error("you must specify input and output file names")
    exit_status = ES_OPTIONS_ERR
  # end if
  sys.exit(exit_status)
# end if
