from annotations import ensembl, knownGene, refGene, aceview, ensg
from utilities import track, assembly
from utilities import tools, intspan
from optparse import OptionParser
import ConfigParser, os, sys, re
import event
import pprint

pp = pprint.PrettyPrinter(indent=0)

class MatchResult:

    first = 0
    last = 0
    middle = 0
    align_id = None
    txt_id = None
    align = None
    txt = None
    coord = ""
    matched_blocks = None
    matched_exons = None
    events = None
    num_blocks = 0
    num_exons = 0
    model = None
    model_weight = None
    coverage = None

    def __init__(self, align_id, txt_id, num_blocks, num_exons, align=None, model=None, txt=None):
        self.align_id = align_id
        self.txt_id = txt_id
        self.num_blocks = num_blocks
        self.num_exons = num_exons
        self.align = align
        self.txt = txt
        self.model = model
        self.matched_blocks = []
        self.matched_exons = []
        self.events = []
        self.coverage = 0.0

    def score(self):
        score = float(self.middle) + float(self.first) + float(self.last)

        for e in self.events:
            if e.event_type == "AS5" or e.event_type == "AS3":
                score += 0.5
            elif e.event_type == "AS53":
                score += 1.0
        
        return score

    def set_coverage(self, exon_bases, txt_len):
        self.coverage = float(exon_bases)/float(txt_len)

    def compare_matches(self, m1, m2):
        if m1.score() > m2.score():
            return -1
        elif m1.score() < m2.score():
            return 1
        elif m1.num_blocks == m1.num_exons and m2.num_blocks != m2.num_exons:
            return -1
        elif m1.num_blocks != m1.num_exons and m2.num_blocks == m2.num_exons:
            return 1
        elif int(m1.middle) > int(m2.middle):
            return -1
        elif int(m1.middle) < int(m2.middle):
            return 1
        else:
            if int(m1.first) or int(m1.last) or int(m2.first) or int(m2.last):
                if int(m1.first) + int(m1.last) > int(m2.first) + int(m2.last):
                    return -1
                elif int(m1.first) + int(m1.last) < int(m2.first) + int(m2.last):
                    return 1
                elif m1.events or m2.events:
                    if m1.events and not m2.events:
                        return 1
                    elif not m1.events and m2.events:
                        return -1
                    elif len(m1.events) < len(m2.events):
                        return -1
                    elif len(m2.events) < len(m1.events):
                        return 1
                    else:
                        if m1.model_weight > m2.model_weight:
                            return -1
                        elif m1.model_weight < m2.model_weight:
                            return 1
                        return 0

                else:
                    if m1.model_weight > m2.model_weight:
                        return -1
                    elif m1.model_weight < m2.model_weight:
                        return 1
                    return 0
                    
            else:
                if m1.events and not m2.events:
                    return 1
                elif not m1.events and m2.events:
                    return -1
                elif len(m1.events) < len(m2.events):
                    return -1
                elif len(m2.events) < len(m1.events):
                    return 1
                else:
                    if m1.model_weight > m2.model_weight:
                        return -1
                    elif m1.model_weight < m2.model_weight:
                        return 1
                    return 0
                

    def full_match(self):
        if self.num_blocks == len(self.matched_blocks) and self.num_exons == len(self.matched_blocks) and int(self.middle) == (self.num_exons-2) * 2 and int(self.first) > 0 and int(self.last) > 0 and len(self.events) == 0:
            return True
        else:
            return False
          
    def report(self, details=False):
        report = ""
        
        if not self.matched_blocks:
            report = "%s no_match against %s total_blocks=%d\n" % (self.align_id, self.txt_id, self.num_blocks)
        else:
            if self.full_match():
                status = "full_match"
            else:
                status = "partial_match"
            report = "%s matches %s model:%s(wt:%d) in %d blocks total_blocks=%d total_exons=%d %s coord:%s score:%.1f events:%d coverage:%s\n" % (self.align_id, self.txt_id, self.model, self.model_weight, len(self.matched_blocks), self.num_blocks, self.num_exons, status, self.coord, self.score(), len(self.events), "%.3f" % (self.coverage))
            if details:
                for i in range(len(self.matched_blocks)):
                    report += "block %d = exon %d\n" % (self.matched_blocks[i], self.matched_exons[i])

        return report

class ModelMatcher:
    
    aligns = None
    report_all = True
    config_file = "/home/rchiu/work/wtss/wtss/gene_models.cfg"
    config = None
    model_order = []

    #path where annotation files (gene models, conservation, repeats) are
    annot_path = None
    #key=gene_models,conservation,repeat value=full path of data file(directory for last 2)
    annots = {}

    full_matches = None
    partial_matches = None
    non_matches = None

    log_file = None
    log = None
    debug = None

    def __init__(self, aligns, config_file=None, species=None, model_order=[], log_file=None, debug=False):
        self.aligns = aligns
        self.config_file = config_file
        self.species = species
        self.model_order = model_order
        self.log_file = log_file
        self.debug = debug
        self.full_matches = {}
        self.partial_matches = {}
        self.non_matches = {}
        #self.feature_matches = {}

        if self.log_file:
            self.log = open(self.log_file, 'w')

        if self.config_file and self.species:
            self.get_models()

        self.splice_motives = self.get_splice_motives()


    def get_models(self):
        "get gene model info from config file info"
        
        self.config = ConfigParser.ConfigParser()
        self.config.read(self.config_file)

        if self.species in self.config.sections():
            for field in self.config.options(self.species):
                value = self.config.get(self.species, field)

                if field == "order" and not self.model_order:
                    self.model_order = value.split(',')
                elif field == "path":
                    self.annot_path = value
                elif field == "splice":
                    self.splice_file = value
                else:
                    self.annots[field] = value

            if self.annots:
                for m,f in self.annots.iteritems():
                    full_path = self.annot_path + "/" + f
                    if self.log:
                        self.log.write(m + ":" + full_path + "\n")
                    if os.path.exists(full_path):
                        self.annots[m] = full_path

        if self.log:
            self.log.write("model order: %s\n" % ','.join(self.model_order))

    def get_splice_motives(self):
        "get canonical splice motives"

        motives = {}

        if os.path.isfile(self.splice_file):
            for line in open(self.splice_file, 'r'):
                cols = line.rstrip("\n").split(' ')
                if len(cols) == 2:
                    motives[cols[0]] = cols[1]

        if len(motives.keys()) > 0:
            return motives
        else:
            return None
        

    def match_models(self):
        all_txts = []
        for i in range(len(self.model_order)):
            model = self.model_order[i]
            model_file = self.annots[model]
            
            if not self.annots.has_key(model) or not os.path.exists(model_file):
                if self.log:
                    self.log.write("doesn't have model: %s %s\n" % (model, model_file))
                continue

            txts = {
                'e': ensembl.parse,
                'r': refGene.parse,
                'k': knownGene.parse,
                'a': aceview.parse,
                'x': ensg.parse
                }[model](model_file)

            #set model in transcript for sorting
            for txt in txts:
                txt.model = model
                txt.weight = len(self.model_order) - i

            all_txts.extend(txts)

        matches = self.compare_model(self.aligns, all_txts)

        #full matches
        full_matches = {}
        for m in matches:
            if m.full_match():
                if not full_matches.has_key(m.align.query):
                    full_matches[m.align.query] = [m]
                else:
                    full_matches[m.align.query].append(m)

        #sort full matches (full matches may have alt.splicing events)
        for contig in full_matches.keys():
            contig_matches = full_matches[contig]
            contig_matches.sort(matches[0].compare_matches)

            self.full_matches[contig] = contig_matches

        #partial matches
        partial_matches = {}
        for m in matches:
            if not self.full_matches.has_key(m.align.query) and len(m.matched_blocks) > 0:
                if not partial_matches.has_key(m.align.query):
                    partial_matches[m.align.query] = [m]
                else:
                    partial_matches[m.align.query].append(m)

        #sort partial matches
        for contig in partial_matches.keys():
            contig_matches = partial_matches[contig]
            contig_matches.sort(matches[0].compare_matches)

            self.partial_matches[contig] = contig_matches

        #non matches
        for m in matches:
            if not self.full_matches.has_key(m.align.query) and not self.partial_matches.has_key(m.align.query):
                if not self.non_matches.has_key(m.align.query):
                    self.non_matches[m.align.query] = [m]
                else:
                    self.non_matches[m.align.query].append(m)
        
            
        if self.log:
            self.log.write("end aligns: %d(full) %d(partial) %d(non)\n" % (len(self.full_matches.keys()), len(self.partial_matches.keys()), len(self.non_matches.keys())))


    def output_coverage(self, output_file):
        "output coverage data"

        ff = open(output_file, 'w')
        match_by_txt = {}

        for matches in self.full_matches.values():
            for match in matches:
                if not match_by_txt.has_key(match.txt_id):
                    match_by_txt[match.txt_id] = [match]
                else:
                    match_by_txt[match.txt_id].append(match)
        
        for matches in self.partial_matches.values():
            for match in matches:
                if not match_by_txt.has_key(match.txt_id):
                    match_by_txt[match.txt_id] = [match]
                else:
                    match_by_txt[match.txt_id].append(match)

        for txt_id in sorted(match_by_txt.keys()):
            all_blocks = []

            match_by_txt[txt_id].sort(lambda x,y: int(y.coverage*1000-x.coverage*1000))
            for match in match_by_txt[txt_id]:
                all_blocks.append(match.align.blocks)

            union_blocks = intspan.union(all_blocks)

            coverage_sum = intspan.intersect(match_by_txt[txt_id][0].txt.exons, union_blocks)

            #print "testing", match_by_txt[txt_id][0].align.contig.num, match_by_txt[txt_id][0].align.contig.length, match_by_txt[txt_id][0].align.contig.coverage, match_by_txt[txt_id][0].align.contig.normalized_coverage()

            line = "\t".join((txt_id, str(coverage_sum), str(match_by_txt[txt_id][0].txt.length), match_by_txt[txt_id][0].align_id, "%.3f" % (match_by_txt[txt_id][0].coverage), str(len(match_by_txt[txt_id])), "%.3f" % (float(coverage_sum)/float(match_by_txt[txt_id][0].txt.length)), match_by_txt[txt_id][0].align.contig.normalized_coverage()))
            ff.write(line + "\n")

        ff.close()

    def output_mapping(self, output_file):
        "output mapping to all models"
        
        ff = open(output_file, 'w')
        for contig, matches in self.full_matches.iteritems():
            for match in matches:
                ff.write(match.report())

        for matches in self.partial_matches.values():
            for match in matches:
                ff.write(match.report())

        for contig, matches in self.non_matches.iteritems():
            ff.write(contig + " no_match " + str(len(matches[0].align.blocks)) + "\n")
            
        ff.close()
                    
    def group_events(self):
        events = {}

        for contig in sorted(self.full_matches.keys()):
            best_match = self.full_matches[contig][0]

            if best_match.events:
                for e in best_match.events:
                    if e.novel and not e.artefact:
                        coord = e.coordinate()
                        if not events.has_key(e.event_type):
                            events[e.event_type] = {coord:[e]}
                        else:
                            if not events[e.event_type].has_key(coord):
                                events[e.event_type][coord] = [e]
                            else:
                                events[e.event_type][coord].append(e)

        for contig in sorted(self.partial_matches.keys()):
            best_match = self.partial_matches[contig][0]

            if best_match.events:
                for e in best_match.events:
                    #print e.event_type, e.novel, e.artefact, e.coordinate()
                    if e.novel and not e.artefact:
                        coord = e.coordinate()
                        if not events.has_key(e.event_type):
                            events[e.event_type] = {coord:[e]}
                        else:
                            if not events[e.event_type].has_key(coord):
                                events[e.event_type][coord] = [e]
                            else:
                                events[e.event_type][coord].append(e)

        for contig in sorted(self.non_matches.keys()):
            for match in self.non_matches[contig]:
                if match.events:
                    for e in match.events:
                        coord = e.coordinate()
                        if not events.has_key(e.event_type):
                            events[e.event_type] = {coord:[e]}
                        else:
                            if not events[e.event_type].has_key(coord):
                                events[e.event_type][coord] = [e]
                            else:
                                events[e.event_type][coord].append(e)

        return events

    def output_events(self, output_file, bed_file=None):
        events = self.group_events()
        
        output = open(output_file, 'w')

        count1 = 1
        for event_type in sorted(events):
            for coord in events[event_type]:
                count2 = 1
                for e in events[event_type][coord]:
                    if len(events[event_type][coord]) == 1:
                        count = count1
                    else:
                        count = "%d.%d" % (count1, count2)
                    
                    output.write(str(count) + " " + e.details() + "\n")

                    count2 += 1

                count1 += 1

        output.close()

        if bed_file:
            self.output_bed(events, bed_file)

    def output_bed(self, events, output_file):
        event_types = ['novel_exon', 'AS53', 'skipped_exon', 'retained_intron', 'novel_utr', 'novel_intron', 'novel_transcript', 'AS5', 'AS3']
        priority_base = 0.00001
        priority_increment = 0.000001
        colors = ['178,34,34', '178,34,34', '30,144,255', '0,128,0', '105,105,105', '72,61,139', '128,0,128', '255,140,0',  '255,140,0']

        output = open(output_file, 'w')
        for i in range(len(event_types)):
            event_type = event_types[i]
            color = colors[i]
            priority = priority_base + i * priority_increment

            if events.has_key(event_type):
                output.write("track name=%s description=%s priority=%.6f color=%s visibility=pack\n" % (event_type, event_type, priority, color))
            
                for coord in events[event_type]:
                    chrom, start, end = re.split('[:-]', coord)
                    bed_start = str(int(start)-1)
                
                    for e in events[event_type][coord]:
                        line = "\t".join([chrom, bed_start, end, e.contig])
                        output.write(line + "\n")
                    
        output.close()
    
        
    def compare_model(self, aligns, txts):
        #best match for each contig
        best_matches = []

        #sort alignments(aligns) and transcripts(txts) by chr
        aligns_by_chr = {}
        for a in aligns:
            if not a.target in aligns_by_chr:
                aligns_by_chr[a.target] = [a]
            else:
                aligns_by_chr[a.target].append(a)

        
        txts_by_chr = {}
        for t in txts:
            if not t.chrom in txts_by_chr:
                txts_by_chr[t.chrom] = [t]
            else:
                txts_by_chr[t.chrom].append(t)
        
        #go through transcript for each contig
        #for chrom, aligns in sorted(aligns_by_chr.iteritems(), key = sort):
        for chrom, aligns in aligns_by_chr.iteritems():
            if not chrom in txts_by_chr.keys():
                continue

            #sort alignments by start position
            aligns.sort(lambda x,y: int(x.tstart)-int(y.tstart))

            #sort transcripts by start position - don't do because want to process transcripts in user-specified order
            txts = txts_by_chr[chrom]
            #txts.sort(lambda x,y: int(x.txStart)-int(y.txStart))

            for align in aligns:

                #orf prediction for all alignments if sequence available
                orf = None
                orfs = []
                if align.contig:
                    orf = event.best_orf(align.contig.sequence)
                    #orfs = event.find_orfs2(align.contig.sequence, 1, align.orient)
                    
                matches = []
                overlapping_txts = []
                
                for txt in txts:
                    #starts matching when alignment and transcript overlaps
                    if intspan.overlap([align.tstart, align.tend], [txt.txStart, txt.txEnd]):
                        overlapping_txts.append(txt)
                        
                        result = self.match_exons(align.query, txt.full_name(), align.blocks, txt.exons, txt.chrom)
                        result.model = txt.model
                        result.model_weight = txt.weight
                        result.txt = txt
                        result.align = align

                        for e in result.events:
                            e.match_result = result

                        #get coverage
                        exon_bases = intspan.intersect(align.blocks, txt.exons)
                        result.set_coverage(exon_bases, txt.length)

                        matches.append(result)

                if matches:
                    matches.sort(matches[0].compare_matches)

                    #no match - introduce novel transcript event
                    if not matches[0].matched_blocks and len(align.blocks) >= 3:
                        e = event.Event("novel_transcript")
                        e.contig = align.query
                        e.align = align
                        e.chrom = chrom
                        e.align_coords = align.blocks
                        if orf:
                            e.test_orf(orf)
                            
                        matches[0].events.append(e)
           
                    for m in matches:
                        m.coord = "%s:%s-%s" % (align.target, align.tstart, align.tend)
                        #m.align = align

                        if m.events:
                            #test for novelty among models
                            novel_events = []
                            for e in m.events:
                                novel_events.extend(e.set_novelty(overlapping_txts, matches))
                                e.splice_motives = self.splice_motives

                            if novel_events:
                                for ne in novel_events:
                                    variant = event.Event(ne['event_type'])
                                    variant.match_result = m
                                    tools.set_attrs(variant, ne)
                                    m.events.append(variant)

                            #test for validity of AS
                            novel_as5 = [e for e in m.events if e.novel and e.event_type == "AS5"]
                            novel_as3 = [e for e in m.events if e.novel and e.event_type == "AS3"]
                            as3_dict = {}
                            for e in novel_as3:
                                if not as3_dict.has_key(e.transcript):
                                    as3_dict[e.transcript] = {}
                                for exon in e.exons:
                                    as3_dict[e.transcript][int(exon)] = e

                            for e in novel_as5:
                                for exon in e.exons:
                                    if as3_dict.has_key(e.transcript) and as3_dict[e.transcript].has_key(int(exon)-1):
                                        as3 = as3_dict[e.transcript][int(exon)-1]
                                        
                                        e.artefact = True
                                        as3.artefact = True

                            #test using orf
                            novel_seqs = [e for e in m.events if e.novel and e.event_type == "retained_intron" or e.event_type == "novel_exon" or e.event_type == "novel_intron" or e.event_type == "novel_utr" or e.event_type == "skipped_exon" or e.event_type == "AS5" or e.event_type == "AS3" or e.event_type == "AS53"]
                            if orf:
                                for e in novel_seqs:
                                    e.test_orf(orf)
                                        
                        best_matches.append(m)

                #no trancript overlapping, keep a result for no match
                else:
                    result = MatchResult(align.query, txts[0].name, len(align.blocks), len(txt.exons), align)
                    
                    if len(align.blocks) >= 3:
                        e = event.Event("novel_transcript")
                        e.contig = align.query
                        e.align = align
                        e.chrom = chrom
                        e.align_coords = align.blocks
                        if orf:
                            e.test_orf(orf)
                        result.events.append(e)
                    
                    best_matches.append(result)

                        

        return best_matches


    def match_exons(self, align_id, txt_id, blocks, exons, chrom=None):
        if self.log:
            self.log.write("%s %d:\t%s\n" % (align_id, len(blocks), pp.pformat(blocks).replace("\n", " ")))
            self.log.write("%s %d:\t%s\n" % (txt_id, len(exons), pp.pformat(exons).replace("\n", " ")))

        result = MatchResult(align_id, txt_id, len(blocks), len(exons))

        #find start (first overlap of alignment block and exon) first
        b = 0
        e = 0
        b_start = None
        e_start = None

        #identify starting matching block 0 -> when they overlap
        while b < len(blocks):
            block = blocks[b]

            while e < len(exons):
                exon = exons[e]

                #if not overlap, on to next exon
                if not intspan.overlap(block, exon):
                    e += 1
                    continue
                
                #last block (but not only block) overlap first exon, not real, skip
                elif b == len(blocks)-1 and e == 0 and len(blocks) > 1:
                    break

                #last exon overlap first block (but not only block), not real, skip
                elif e == len(exons)-1 and b == 0 and len(blocks) > 1:
                    break

                #start identified, done get out of exon loop
                b_start = b
                e_start = e
                break

            #start not identified, next alignment block
            if b_start == None and e_start == None:
                e = 0
                b += 1
            #start identified, done get out of block loop
            else:
                break
        
        if b_start >= 0 and e_start >= 0:
            b = b_start
            e = e_start

            if self.log:
                self.log.write("start %d %d %d %d\n" % (b, e, len(blocks), len(exons)))

            if b_start > 0 and e_start == 0:
                if self.log:
                    self.log.write("novel 5utr %s %s %s-%s\n" % (align_id, txt_id, blocks[0][0], blocks[-1][-1]))

                variant = event.Event("novel_utr")
                tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                tools.set_attrs(variant, {'align_blocks':range(1,b_start+1), 'align_coords':blocks[0:b_start]})
                tools.set_attrs(variant, {'exons':[1], 'exon_coords':[exons[0]]})
                result.events.append(variant)

            while b < len(blocks):
                block = blocks[b]
                        
                while e < len(exons):
                    exon = exons[e]

                    matched = self.match(blocks, exons, b, e)
                    if self.log:
                        self.log.write("%d %d %s %s %s\n" % (b, e, pp.pformat(block), pp.pformat(exon), matched))
                        
                    if matched and not matched == 'sync':
                        result.matched_exons.append(e+1)
                        result.matched_blocks.append(b+1)

                        if not matched == 'perfect' and len(blocks) > 1:
                            #not e == 0: not considering first exon (start 5'utr)
                            #not (b == 0 and int(block[0]) > int(exon[0])): first block and overshoot and overridden by last condition
                            #not (int(exon[0]) < int(blocks[b-1][-1])): start of exon not going over to previous block - not novel intron
                            #not (int(block[0]) < int(exons[e-1][-1])): start of block not going over to previous exon - not retained intron
                            #not b == 0: not considering first block (overrides 2nd condition -- temporarily turned off, because see a real extension of an internal exon
                            if matched == 'as5' and not e == 0 and not b == 0 and not (b == 0 and int(block[0]) > int(exon[0])) and not (b > 0 and int(exon[0]) < int(blocks[b-1][-1])) and not (e > 0 and int(block[0]) < int(exons[e-1][-1])):
                                if self.log:
                                    self.log.write("5as %s %s %s-%s\n" % (align_id, txt_id, block[0], block[1]))

                                variant = event.Event("AS5")
                                tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                                tools.set_attrs(variant, {'align_blocks':[b+1], 'align_coords':blocks[b]})
                                tools.set_attrs(variant, {'exons':[e+1], 'exon_coords':[exons[e]]})
                                result.events.append(variant)

                            #not e == len(exons)-1: not considering last exon (end 3'utr)
                            #not (b == len(blocks)-1 and int(block[-1]) < int(exon[-1])): last block overshoot an internal exon - overridden by last condition
                            #not b == len(blocks)-1: not considering last block: overrides 2nd condition -- temporarily turned off
                            if matched == 'as3' and not e == len(exons)-1 and not b == len(blocks)-1:
                                if self.log:
                                    self.log.write("3as %s %s %s-%s\n" % (align_id, txt_id, block[0], block[1]))

                                variant = event.Event("AS3")
                                tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                                tools.set_attrs(variant, {'align_blocks':[b+1], 'align_coords':blocks[b]})
                                tools.set_attrs(variant, {'exons':[e+1], 'exon_coords':[exons[e]]})
                                result.events.append(variant)

                            if matched == 'as53' and not e == 0 and not b == 0 and not (b == 0 and int(block[0]) > int(exon[0])) and not (b > 0 and int(exon[0]) < int(blocks[b-1][-1])) and not (e > 0 and int(block[0]) < int(exons[e-1][-1])) and not e == len(exons)-1 and not b == len(blocks)-1:
                                if self.log:
                                    self.log.write("as53 %d %d\n" % (b, b+1))

                                variant = event.Event("AS53")
                                tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                                tools.set_attrs(variant, {'align_blocks':[b+1], 'align_coords':[blocks[b]]})
                                tools.set_attrs(variant, {'exons':[e+1], 'exon_coords':[exons[e]]})
                                result.events.append(variant)

                        #record matches = CHECK THIS!!!!!!!!!!
                        if e == 0 or e == len(exons)-1:
                            if matched == 'perfect':
                                score = 2
                            #'as53' will have no score
                            elif matched == 'as5' or matched == 'as3':
                                score = 1
                            else:
                                score = 0

                            if e == 0:
                                result.first = score

                            else:
                                result.last = score

                        else:
                            if matched == 'perfect':
                                result.middle += 2

                            #start(5') is different, end(3') is the same
                            elif matched == 'as5' and b == 0:
                                result.middle += 1

                            #start(5') is the same, end(3') is different
                            elif matched == 'as3' and b == len(blocks)-1:
                                result.middle += 1

                        b += 1
                        e += 1
                        break

                    #novel exon
                    elif int(block[1]) < int(exon[0]):
                        #increment block, until finding matching block/exon
                        bb = b + 1

                        matched = False
                        while (not matched or matched == 'sync') and bb <= len(blocks)-1:
                            matched = self.match(blocks, exons, bb, e)
                            if not matched or matched == 'sync':
                                bb += 1

                        #if bb == len(blocks)-1, check if loop stops because of match or last index
                        #if loops stops because of finding matching exon, then novel exons is between b and bb
                        if matched:
                            if self.log:
                                self.log.write("novel exons %d %d\n" % (b, bb))

                            variant = event.Event("novel_exon")
                            tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                            tools.set_attrs(variant, {'align_blocks':range(b+1,bb+1), 'align_coords':blocks[b:bb]})
                            #tools.set_attrs(variant, {'exons':[e], 'exon_coords':[exons[e-1]]})
                            tools.set_attrs(variant, {'exons':[e,e+1], 'exon_coords':[exons[e-1:e+1]]})
                            result.events.append(variant)
 
                        #set next comparison to be bb, if it's the last index, then handle it outside??, it will handle alt.splicing
                        b = bb
                        break

                    #skipped exon
                    elif int(block[0]) > int(exon[1]):
                        #increment exon, until finding matching block/exon
                        ee = e + 1
                        
                        matched = False
                        while (not matched or matched == 'sync') and ee <= len(exons)-1:
                            matched = self.match(blocks, exons, b, ee)
                            if not matched or matched == 'sync':
                                ee += 1

                        #requires ee to be matched
                        if matched:                       
                            if self.log:
                                self.log.write("skipped exons %d %d\n" % (e, ee))

                            variant = event.Event("skipped_exon")
                            tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                            #tools.set_attrs(variant, {'align_blocks':[b], 'align_coords':[blocks[b-1]]})
                            tools.set_attrs(variant, {'align_blocks':range(b,b+2), 'align_coords':blocks[b-1:b+1]})
                            tools.set_attrs(variant, {'exons':range(e+1,ee+1), 'exon_coords':exons[e:ee]})
                            result.events.append(variant)

                        e = ee
                        break

                    #retained intron
                    elif int(block[1]) > int(exon[1]):
                        #increment exon, until finding matching block/exon
                        ee = e + 1

                        if ee <= len(exons)-1 and int(exons[ee][0]) < int(block[1]):
                            matched = False
                            while (not matched or matched == 'sync') and ee <= len(exons)-1:
                                matched = self.match(blocks, exons, b, ee)
                                if not matched or matched == 'sync':
                                    ee += 1

                                if matched and not matched == "as3":
                                    if self.log:
                                        self.log.write("retained intron %d %d\n" % (e, ee))

                                    variant = event.Event("retained_intron")
                                    tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                                    tools.set_attrs(variant, {'align_blocks':[b+1], 'align_coords':[blocks[b]]})
                                    tools.set_attrs(variant, {'exons':range(e+1,ee+2), 'exon_coords':exons[e:ee+1]})
                                    result.events.append(variant)
                        else:
                            b = b + 1

                        e = ee
                        break

                    #novel intron/indel
                    elif int(block[1]) < int(exon[1]):
                        #increment block, until finding matching block/exon
                        bb = b + 1

                        if bb <= len(blocks)-1 and int(blocks[bb][0]) < int(exon[1]):
                            matched = False
                            while (not matched or matched == 'sync') and bb <= len(blocks)-1:
                                matched = self.match(blocks, exons, bb, e)
                                if not matched or matched == 'sync':
                                    bb += 1

                            if matched:
                                if self.log:
                                    self.log.write("novel intron %d %d\n" % (b, bb))

                                variant = event.Event("novel_intron")
                                tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                                tools.set_attrs(variant, {'align_blocks':range(b+1,bb+2), 'align_coords':blocks[b:bb+1]})
                                tools.set_attrs(variant, {'exons':[e+1], 'exon_coords':[exons[e]]})
                                result.events.append(variant)
                        else:
                            e = e + 1
                            
                        b = bb
                        break

                    else:
                        if self.log:
                            self.log.write("??\n")
                        e += 1
                        break

                    break

                #make sure the loop ends
                if e > len(exons)-1:
                    break

            #if last matched/synced is not last block and is last exon
            if b < len(blocks)-1 and e == len(exons)-1:
                if self.log:
                    self.log.write("novel 3utr %s %s %d %d %s-%s" % (align_id, txt_id, b, e, exons[0][0], exons[-1][-1]))
                    
                variant = event.Event("novel_utr")
                tools.set_attrs(variant, {'contig':align_id, 'transcript':txt_id, 'chrom':chrom})
                tools.set_attrs(variant, {'align_blocks':range(b+2,len(blocks)+1), 'align_coords':blocks[b+1:]})
                tools.set_attrs(variant, {'exons':[len(exons)], 'exon_coords':[exons[-1]]})
                result.events.append(variant)

        else:
            if self.log:
                self.log.write("do not overlap could not find start")

        if self.log:
            self.log.write("----\n")

        if result.events and not result.matched_blocks:
            del result.events[:]
                            
        return result

        
    def match(self, blocks, exons, b, e):
        match = False

        if int(blocks[b][0]) == int(exons[e][0]) and int(blocks[b][1]) == int(exons[e][1]):
            match = "perfect"

        #single block and embedded within intron
        elif len(blocks) == 1:
            #one block embedded within an exon
            if intspan.subsume(blocks[b], exons[e]):
                match = "perfect"
            elif int(blocks[b][1]) > int(exons[e][1]):
                match = "as3"
            elif not e == 0 and int(blocks[b][0]) < int(exons[e][0]):
                match = "as5"

        elif e == len(exons)-1:
            #if not overlap, or cut into next exon return false
            if intspan.overlap(blocks[b], exons[e]):
                #last block, last exon
                if b == len(blocks)-1:
                    if int(blocks[b][0]) == int(exons[e][0]):
                        match = "perfect"
                    else:
                        match = "as5"
                #not last block, novel exon
                else:
                    #2nd condition check if it has retained intron
                    if int(blocks[b][0]) == int(exons[e][0]) or int(blocks[b][0]) < int(exons[e-1][1]):
                        match = "as3"
                    elif int(blocks[b][1]) == int(exons[e][1]):
                        match = "as5"
                    else:
                        match = "as53"

        elif b == len(blocks)-1:
            #if not overlap, or cut into next exon return false
            #first check overlap
            if intspan.overlap(blocks[b], exons[e]):
                #last block vs middle exon
                if e < len(exons)-1:
                    #not retained intron
                    if int(blocks[b][1]) < int(exons[e+1][0]):
                        if int(blocks[b][0]) == int(exons[e][0]) or int(blocks[b][0]) < int(exons[e-1][1]):
                            match = "as3"
                        elif int(blocks[b][-1]) == int(exons[e][-1]):
                            match = "as5"
                        else:
                            match = "as53"
                #last block vs last exon (last exon = only check as5)
                elif not (int(blocks[b][0]) == int(exons[e][0])):
                    match = "as5"
                #perfect = 3' end not match
                else:
                    match = "perfect"

        #as5 or as3
        elif int(blocks[b][1]) < int(exons[e+1][0]) and int(exons[e][1]) < int(blocks[b+1][0]) and (int(blocks[b][0]) == int(exons[e][0]) or int(blocks[b][1]) == int(exons[e][1])):
            if int(blocks[b][0]) == int(exons[e][0]):
                match = "as3"
            else:
                match = "as5"
                    
        elif int(blocks[b][1]) < int(exons[e+1][0]) and int(exons[e][1]) < int(blocks[b+1][0]) and int(exons[e+1][0]) < int(blocks[b+1][1]) and int(blocks[b+1][0]) < int(exons[e+1][1]):
            if int(blocks[b][0]) == int(exons[e][0]):
                match = "as3"
            elif int(blocks[b][1]) == int(exons[e][1]):
                match = "as5"
            else:
                match = "as53"
                
        #move on
        elif int(blocks[b][0]) > int(exons[e][1]):
            match = "sync"

        return match


if __name__ == '__main__':
    usage = "Usage: %prog alignment_file"

    parser = OptionParser(usage=usage)
    parser.add_option("-f", "--config", dest="config", help="config file", default="/home/rchiu/work/wtss/wtss/gene_models.cfg")
    parser.add_option("-s", "--species", dest="species", help="species e.g. human, mouse")
    parser.add_option("-e", "--ensembl", dest="ensembl", help="ensembl", action="store_true", default=False)
    parser.add_option("-x", "--ensg", dest="ensg", help="ensembl gene", action="store_true", default=False)
    parser.add_option("-k", "--knowngenes", dest="knowngenes", help="known genes", action="store_true", default=False)
    parser.add_option("-r", "--refgenes", dest="refgenes", help="refseq genes", action="store_true", default=False)
    parser.add_option("-a", "--aceview", dest="aceview", help="aceview", action="store_true", default=False)
    parser.add_option("-m", "--models", dest="models", help="model order e.g. k,e,r,a")
    parser.add_option("-t", "--track", dest="track", help="track_file track_name", nargs=2)
    parser.add_option("-v", "--events", dest="events", help="event output file")
    parser.add_option("-l", "--log", dest="log", action="store_true", default=False)
    parser.add_option("-d", "--debug", dest="debug", action="store_true", default=False)
    parser.add_option("-o", "--outdir", dest="outdir", help="output directory")
    parser.add_option("-C", "--chr", dest="chr", help="chromosome")
    parser.add_option("-S", "--sequence", dest="fasta", help="fasta file")
    parser.add_option("-z", "--coverage", dest="coverage", help="coverage file", action="store_true", default=False)
    parser.add_option("-K", "--kmer", dest="k", help="k-mer length")
    parser.add_option("-b", "--bed", dest="bed", help="bed file for events", action="store_true", default=False)

    (options, args) = parser.parse_args()

    if len(args) <= 1:
        models = []
        if options.ensembl:
            models = ['e']
        elif options.knowngenes:
            models = ['k']
        elif options.refgenes:
            models = ['r']
        elif options.aceview:
            models = ['a']
        elif options.ensg:
            models = ['x']
        if not models and options.models:
            models = options.models.split(',')

        #get alignments from track
        trk = track.Track(args[0])

        ext = os.path.splitext(args[0])[1]

        aligns = []
        if ext == ".gff":
            aligns = trk.parse_gff()
        elif ext == ".psl":
            aligns = trk.parse_psl()

        #filter chrom-specific alignments
        if options.chr:
            if options.chr[:3].lower() == 'chr':
                chrom = options.chr.lower()
            else:
                chrom = 'chr' + options.chr

            aligns = [a for a in aligns if a.target == chrom]

        log_file = None
        if options.outdir:
            if not os.path.isdir(options.outdir):
                os.mkdir(options.outdir)

            if options.log:
                log_file = options.outdir + "/log.txt"

        #get contig sequences
        contigs = None
        if options.fasta:
            sys.stderr.write("getting contig sequences...")
            ass = assembly.Assembly(0, fasta=options.fasta)
            contigs = ass.get_contigs(sequence=True)
            sys.stderr.write("done\n")
            
            if contigs:
                contig_dict = dict((c.num, c) for c in contigs)
                #print len(contig_dict.keys())

                for align in aligns:
                    if contig_dict.has_key(align.query):
                        align.contig = contig_dict[align.query]

                        if options.k:
                            align.contig.k = options.k
                        elif "k" in align.contig.num:
                            align.contig.k = align.contig.num.split(':')[0].lstrip("k")
                            
        mm = ModelMatcher(aligns, options.config, options.species, log_file=log_file, debug=options.debug)

        if models:
            mm.model_order = models

        #funnel through transcript models
        if mm.model_order:
            mm.match_models()

        if options.coverage:
            mm.output_coverage(options.outdir + "/coverage.txt")

        if options.outdir and os.path.isdir(options.outdir):
            mm.output_mapping(options.outdir + "/mapping.txt")

            bed_file = None
            if options.bed:
                bed_file = options.outdir + "/events.bed"
            mm.output_events(options.outdir + "/events.txt", bed_file)

        if options.track:
            mm.output_track(options.track[0], options.track[1])

    else:
        parser.error("incorrect number of arguments")
