#!/usr/bin/perl

=head1 NAME

extractCutoffPolyA.pl - Extracts PAM reads from a SOLEXA raw reads file


=head1 SYNOPSYS

  % extractCutoffPolyA.pl -f <INPUT> -r <INPUT> -of <OUTPUT> -or <OUTPUT> -m <MODE>[-q <INT> -c <INT>]

=head1 DESCRIPTION

This script reads two SOLEXA read files and extracts pairs which are high in either A or T.
There are 2 modes: select for high proportion of T's or A's.

This script enables inputs to be gzipped or bzip2'ed but requires the library IO::Uncompress::Gunzip or IO::Uncompress::Bunzip2.

Note the read name has appended the number of T's in polyA read.

=head2 NOTES

This is still a work in progress.

=head1 COMMAND-LINE OPTIONS

Options can be abbreviated.  For example, you can use -i for input
    --forward          Forward input from Illumina pipeline
    --reverse          Reverse input from Illumina pipeline
    --of               Output forward FASTQ file
    --or               Output reverse FASTQ file
    --cutoff           Minimum number of A/T required to save
                         Default = 40
    --append           Choose to append the results to the output files
                         Default = 0 (write new)
    --mode             The type of extraction to consider (t or a)
                         Default = "t"
    --append           Append to the output files
                         Default = 0 (stat new)
    --qual             Quality for all bases of the read                                      
                         Default = 5
    
=head1 SEE ALSO

=head1 AUTHOR

Mark Okada, hmokada@sfu.ca

Copyright (c) 2010 Simon Fraser University

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.  See DISCLAIMER.txt for
disclaimers of warranty.

=cut

use strict;
use FindBin;

use lib "$FindBin::RealBin/../lib";

use Data::Dumper;
use Getopt::Long;
use Config::General;
use Switch;
use Data::Dumper;
use vars qw(%OPT %CONF $CUTOFF $TYPE);

#default vals
%OPT = (cutoff => 40, mode => "t", qual => "/1");

GetOptions(\%OPT,
	   "forward=s",
	   "reverse=s",
	   "of=s",
	   "or=s",
	   "append",
	   "cutoff=i",
	   "mode=s",
	   "qual=s");

my $usage = "extractCutoffPolyA.pl -f <INPUT> -r <INPUT> -of <OUTPUT> -or <OUTPUT> -m <MODE>[-q <INT> -c <INT>] ";

# check file type
my ($extention) = $OPT{forward} =~ /\.([^\.]+)$/;

my ($z1, $z2);

if ($extention eq "bz2") {
    use IO::Uncompress::Bunzip2 qw(bunzip2 $Bunzip2Error);
    
    $z1 = new IO::Uncompress::Bunzip2 $OPT{forward}
	or die "bunzip2 failed: $Bunzip2Error\n";
    $z2 = new IO::Uncompress::Bunzip2 $OPT{reverse}
	or die "bunzip2 failed: $Bunzip2Error\n";
} elsif ($extention eq "gz") {
    use IO::Uncompress::Gunzip qw(gunzip $GunzipError);

    $z1 = new IO::Uncompress::Gunzip $OPT{forward}
        or die "gunzip failed: $GunzipError\n";
    $z2 = new IO::Uncompress::Gunzip $OPT{reverse}
        or die "gunzip2 failed: $GunzipError\n";
} else {
    open (F, $OPT{forward}) || die $usage;
    open (R, $OPT{reverse}) || die $usage;
}


if ($OPT{append}) {
    open (OF, ">>". $OPT{of}) || die $usage;
    open (OR, ">>". $OPT{or}) || die $usage;
} else {
    open (OF, ">". $OPT{of}) || die $usage;
    open (OR, ">". $OPT{or}) || die $usage;
}


my @set1;
my @set2;

my $status = 1;
while ($status) {
    
    my ($f, $r);
    
    if ($extention eq "bz2" || $extention eq "gz") {
	my $status1 = $f = $z1->getline();
	my $status2 = $r = $z2->getline();
	$status = 0 unless $status1 && $status2;
    } else {
	my $status1 = $f = <F>;
	my $status2 = $r = <R>;
	$status = 0 unless $status1 && $status2;
    }
    
    proc_set(proc_illum($f),
	     proc_illum($r),
	     *OF,
	     *OR);
}

if ($extention ne "bz2" && $extention ne "gz") {
    close L;
    close R;
}
close OL;
close OR;


# save a FASTQ entry if PAM read: its mate is high A/T
sub proc_set {
    my $f = shift;
    my $r = shift;
    my $fhf = shift;
    my $fhr = shift;
    
    my ($n1,$n2);
    
    if ($OPT{type} eq "a") {
	$n1 = $f->[1] =~ tr/Aa/Aa/;
        $n2 = $r->[1] =~ tr/Aa/Aa/;
    } else {
	$n1 = $f->[1] =~ tr/Tt/Tt/;
	$n2 = $r->[1] =~ tr/Tt/Tt/;
    }
    
    # append the number of A/T's to the read name
    if ($OPT{cutoff} <= $n1) {
	
	# Forward read poly-A/T, save reverse read
	my $read = $r->[0];
	$r->[0] = $read . "_" . $n1 . "/2";
	print $fhr join("\n",@$r),"\n";
    }
    
    if ($OPT{cutoff} <= $n2) {
	# Reverse read poly-A/T, save forward read
	my $read = $f->[0];
	$f->[0] = $read . "_" . $n2 . "/1";
	print $fhf join("\n",@$f),"\n";
    }
}

# convert an illumina raw read formatted file to fastq
sub proc_illum {
    my $input = shift;
    
    my @tok = split "\t", $input;
    
    if ($tok[0] =~ /^([^_]+_\d+)/) {
        $tok[0] = $1;
    } else {
        $tok[0] = $tok[0] . "_" . $tok[1];
    }
    
    my @output = (join(":",("@".$tok[0]),@tok[2..5]),
		  $tok[8],
		  "+",
		  "5" x length($tok[8]));
    return \@output;
}
