#!/usr/bin/perl

=head1 NAME

getrawreadendtag.pl - Reads a raw read file to extract possible end tags

=head1 SYNOPSYS

  % cat RAWREADFILE | getrawreadendtag.pl > FASTQ

=head1 DESCRIPTION

Runs all the scripts for the ...


=head2 NOTES

Read a raw read file from Eland and export a FASTQ read file for those that may be possible end tag files (Nagalakshmi 2009).
Consider only reads in which they are prefixed by consecutive T's for paired end reads, prefixed by T's or suffixed by A's
for single end reads.

This is still a work in progress.

=head1 COMMAND-LINE OPTIONS

Options can be abbreviated.  For example, you can use -i for input
    --nm               Skip NM entries
                         Default = 0
    --a                Consider reads with suffixed A's
    --min              Minimum A or T stretch required to count
                         Default = 5
    --qual             Quality value for the output FASTQ file 
                         Default = 5


=head1 SEE ALSO

=head1 AUTHOR

Mark Okada, hmokada@sfu.ca

Copyright (c) 2009 Simon Fraser University

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.  See DISCLAIMER.txt for
disclaimers of warranty.

=cut

use strict;
use FindBin;

use lib "$FindBin::RealBin/../lib";

use Data::Dumper;
use Getopt::Long;
use Config::General;
use vars qw(%OPT %CONF);

#default vals
%OPT = (qual => 5, min => 5);

GetOptions(\%OPT,
	   "qual=i",
	   "a",
	   "nm",
	   "min=i",
	   "quiet+");




while (<>) {
    chomp;
    my @tok = split "\t";
    
    # skip if not NM (optional)
    next if $OPT{nm} && $tok[10] ne "NM";
    
    
    # number of consecutive A's and T's
    my ($s_lt) = $tok[8] =~ m/^(t*)/i;
    my ($s_ra) = $tok[8] =~ m/(a*)$/i;
    
    
    my $n_lt = length($s_lt);
    my $n_ra = length($s_ra);
    my $score;    
    my $trimmed;
    
    if ($OPT{a} && $n_lt < $n_ra) {
	next if ($n_ra < $OPT{min});
	# this is a poly A end read: XXXXXXXX-AAAA
	
	$score = $n_ra;
	$trimmed = substr($tok[8], 0, length($tok[8])-$n_ra);
	
    } else {
	next if ($n_lt < $OPT{min});
	# this is a poly T end read: TTTT-XXXXXXXX
	
	$score = - $n_lt;
	$trimmed = substr($tok[8], $n_lt);
    }
    
    
    #create name for read - not all reads are consistent
    my $label;
    
    if ($tok[0] =~ /^([^_]+_\d+)/) {
	$label = join(":",($1),@tok[2..5]) . "_$score/$tok[7]";
    } else {
	$label = join(":",($tok[0] . "_" . $tok[1]),@tok[2..5]) . "_$score/$tok[7]";
    }
    
    print join("\n",
	       "@".$label,
	       $tok[8],
	       "+",
	       $OPT{qual} x length($tok[8])),"\n";
    
    
    print join("\n",
	       "@"."tr".$label,
	       $trimmed,
	       "+",
               $OPT{qual} x length($trimmed)),"\n";
    
    
}





1;
